"""Mark scikit-learn classes as configurable"""
import gin

# TODO: Delete when https://github.com/google/gin-config/pull/49 gets merged

# Most of the classes mentioned in the docs:
# https://scikit-learn.org/stable/modules/classes.html

# Clustering
import sklearn.cluster
gin.external_configurable(sklearn.cluster.AffinityPropagation)
gin.external_configurable(sklearn.cluster.AgglomerativeClustering)
gin.external_configurable(sklearn.cluster.Birch)
gin.external_configurable(sklearn.cluster.DBSCAN)
gin.external_configurable(sklearn.cluster.FeatureAgglomeration)
gin.external_configurable(sklearn.cluster.KMeans)
gin.external_configurable(sklearn.cluster.MiniBatchKMeans)
gin.external_configurable(sklearn.cluster.MeanShift)
gin.external_configurable(sklearn.cluster.OPTICS)
gin.external_configurable(sklearn.cluster.SpectralClustering)
gin.external_configurable(sklearn.cluster.SpectralBiclustering)
gin.external_configurable(sklearn.cluster.SpectralCoclustering)

# Cross Decomposition
import sklearn.cross_decomposition
gin.external_configurable(sklearn.cross_decomposition.CCA)
gin.external_configurable(sklearn.cross_decomposition.PLSCanonical)
gin.external_configurable(sklearn.cross_decomposition.PLSRegression)
gin.external_configurable(sklearn.cross_decomposition.PLSSVD)

# Decomposition
import sklearn.decomposition
gin.external_configurable(sklearn.decomposition.DictionaryLearning)
gin.external_configurable(sklearn.decomposition.FactorAnalysis)
gin.external_configurable(sklearn.decomposition.FastICA)
gin.external_configurable(sklearn.decomposition.IncrementalPCA)
gin.external_configurable(sklearn.decomposition.KernelPCA)
gin.external_configurable(sklearn.decomposition.LatentDirichletAllocation)
gin.external_configurable(sklearn.decomposition.MiniBatchDictionaryLearning)
gin.external_configurable(sklearn.decomposition.MiniBatchSparsePCA)
gin.external_configurable(sklearn.decomposition.NMF)
gin.external_configurable(sklearn.decomposition.PCA)
gin.external_configurable(sklearn.decomposition.SparsePCA)
gin.external_configurable(sklearn.decomposition.SparseCoder)
gin.external_configurable(sklearn.decomposition.TruncatedSVD)

# Discriminant Analysis
import sklearn.discriminant_analysis
gin.external_configurable(sklearn.discriminant_analysis.LinearDiscriminantAnalysis)
gin.external_configurable(sklearn.discriminant_analysis.QuadraticDiscriminantAnalysis)

# Ensemble
import sklearn.ensemble
gin.external_configurable(sklearn.ensemble.AdaBoostClassifier)
gin.external_configurable(sklearn.ensemble.AdaBoostRegressor)
gin.external_configurable(sklearn.ensemble.BaggingClassifier)
gin.external_configurable(sklearn.ensemble.BaggingRegressor)
gin.external_configurable(sklearn.ensemble.ExtraTreesClassifier)
gin.external_configurable(sklearn.ensemble.ExtraTreesRegressor)
gin.external_configurable(sklearn.ensemble.GradientBoostingClassifier)
gin.external_configurable(sklearn.ensemble.GradientBoostingRegressor)
gin.external_configurable(sklearn.ensemble.IsolationForest)
gin.external_configurable(sklearn.ensemble.RandomForestClassifier)
gin.external_configurable(sklearn.ensemble.RandomForestRegressor)
gin.external_configurable(sklearn.ensemble.StackingClassifier)
gin.external_configurable(sklearn.ensemble.StackingRegressor)
gin.external_configurable(sklearn.ensemble.VotingClassifier)
gin.external_configurable(sklearn.ensemble.VotingRegressor)

# Feature Extraction
import sklearn.feature_extraction
gin.external_configurable(sklearn.feature_extraction.DictVectorizer)
gin.external_configurable(sklearn.feature_extraction.FeatureHasher)
gin.external_configurable(sklearn.feature_extraction.image.extract_patches_2d)
gin.external_configurable(sklearn.feature_extraction.image.grid_to_graph)
gin.external_configurable(sklearn.feature_extraction.image.img_to_graph)
gin.external_configurable(sklearn.feature_extraction.image.reconstruct_from_patches_2d)
gin.external_configurable(sklearn.feature_extraction.image.PatchExtractor)
gin.external_configurable(sklearn.feature_extraction.text.CountVectorizer)
gin.external_configurable(sklearn.feature_extraction.text.HashingVectorizer)
gin.external_configurable(sklearn.feature_extraction.text.TfidfTransformer)
gin.external_configurable(sklearn.feature_extraction.text.TfidfVectorizer)

# Feature Selection
import sklearn.feature_selection
gin.external_configurable(sklearn.feature_selection.GenericUnivariateSelect)
gin.external_configurable(sklearn.feature_selection.SelectPercentile)
gin.external_configurable(sklearn.feature_selection.SelectKBest)
gin.external_configurable(sklearn.feature_selection.SelectFpr)
gin.external_configurable(sklearn.feature_selection.SelectFdr)
gin.external_configurable(sklearn.feature_selection.SelectFromModel)
gin.external_configurable(sklearn.feature_selection.SelectFwe)
gin.external_configurable(sklearn.feature_selection.RFE)
gin.external_configurable(sklearn.feature_selection.RFECV)
gin.external_configurable(sklearn.feature_selection.VarianceThreshold)
gin.external_configurable(sklearn.feature_selection.chi2)
gin.external_configurable(sklearn.feature_selection.f_classif)
gin.external_configurable(sklearn.feature_selection.f_regression)
gin.external_configurable(sklearn.feature_selection.mutual_info_classif)
gin.external_configurable(sklearn.feature_selection.mutual_info_regression)

# Gaussian Processes
import sklearn.gaussian_process
gin.external_configurable(sklearn.gaussian_process.GaussianProcessRegressor)
gin.external_configurable(sklearn.gaussian_process.GaussianProcessClassifier)
gin.external_configurable(sklearn.gaussian_process.kernels.CompoundKernel)
gin.external_configurable(sklearn.gaussian_process.kernels.ConstantKernel)
gin.external_configurable(sklearn.gaussian_process.kernels.DotProduct)
gin.external_configurable(sklearn.gaussian_process.kernels.ExpSineSquared)
gin.external_configurable(sklearn.gaussian_process.kernels.Exponentiation)
gin.external_configurable(sklearn.gaussian_process.kernels.Hyperparameter)
gin.external_configurable(sklearn.gaussian_process.kernels.Matern)
gin.external_configurable(sklearn.gaussian_process.kernels.PairwiseKernel)
gin.external_configurable(sklearn.gaussian_process.kernels.Product)
gin.external_configurable(sklearn.gaussian_process.kernels.RBF)
gin.external_configurable(sklearn.gaussian_process.kernels.RationalQuadratic)
gin.external_configurable(sklearn.gaussian_process.kernels.Sum)
gin.external_configurable(sklearn.gaussian_process.kernels.WhiteKernel)

# Impute
import sklearn.impute
gin.external_configurable(sklearn.impute.SimpleImputer)
gin.external_configurable(sklearn.impute.MissingIndicator)
gin.external_configurable(sklearn.impute.KNNImputer)

# Isotonic
import sklearn.isotonic
gin.external_configurable(sklearn.isotonic.IsotonicRegression)

# Kernel Approximation
import sklearn.kernel_approximation
gin.external_configurable(sklearn.kernel_approximation.AdditiveChi2Sampler)
gin.external_configurable(sklearn.kernel_approximation.Nystroem)
gin.external_configurable(sklearn.kernel_approximation.RBFSampler)
gin.external_configurable(sklearn.kernel_approximation.SkewedChi2Sampler)

# Kernel Ridge Regression
import sklearn.kernel_ridge
gin.external_configurable(sklearn.kernel_ridge.KernelRidge)

# Linear Classifiers
import sklearn.linear_model
gin.external_configurable(sklearn.linear_model.LogisticRegression)
gin.external_configurable(sklearn.linear_model.LogisticRegressionCV)
gin.external_configurable(sklearn.linear_model.PassiveAggressiveClassifier)
gin.external_configurable(sklearn.linear_model.Perceptron)
gin.external_configurable(sklearn.linear_model.RidgeClassifier)
gin.external_configurable(sklearn.linear_model.RidgeClassifierCV)
gin.external_configurable(sklearn.linear_model.SGDClassifier)

# Classical Linear Regressors
gin.external_configurable(sklearn.linear_model.LinearRegression)
gin.external_configurable(sklearn.linear_model.Ridge)
gin.external_configurable(sklearn.linear_model.RidgeCV)
gin.external_configurable(sklearn.linear_model.SGDRegressor)

# Regressors with variable selection
gin.external_configurable(sklearn.linear_model.ElasticNet)
gin.external_configurable(sklearn.linear_model.ElasticNetCV)
gin.external_configurable(sklearn.linear_model.Lars)
gin.external_configurable(sklearn.linear_model.LarsCV)
gin.external_configurable(sklearn.linear_model.Lasso)
gin.external_configurable(sklearn.linear_model.LassoCV)
gin.external_configurable(sklearn.linear_model.LassoLars)
gin.external_configurable(sklearn.linear_model.LassoLarsCV)
gin.external_configurable(sklearn.linear_model.OrthogonalMatchingPursuit)
gin.external_configurable(sklearn.linear_model.OrthogonalMatchingPursuitCV)

# Bayesian Regressors
gin.external_configurable(sklearn.linear_model.ARDRegression)
gin.external_configurable(sklearn.linear_model.BayesianRidge)

# Multi-task Linear Regressors with Variable Selection
gin.external_configurable(sklearn.linear_model.MultiTaskElasticNet)
gin.external_configurable(sklearn.linear_model.MultiTaskElasticNetCV)
gin.external_configurable(sklearn.linear_model.MultiTaskLasso)
gin.external_configurable(sklearn.linear_model.MultiTaskLassoCV)

# Outlier-robust Regressors
gin.external_configurable(sklearn.linear_model.HuberRegressor)
gin.external_configurable(sklearn.linear_model.RANSACRegressor)
gin.external_configurable(sklearn.linear_model.TheilSenRegressor)

# Miscellaneous
gin.external_configurable(sklearn.linear_model.PassiveAggressiveRegressor)

# Manifold
import sklearn.manifold
gin.external_configurable(sklearn.manifold.Isomap)
gin.external_configurable(sklearn.manifold.LocallyLinearEmbedding)
gin.external_configurable(sklearn.manifold.MDS)
gin.external_configurable(sklearn.manifold.SpectralEmbedding)
gin.external_configurable(sklearn.manifold.TSNE)

# Mixture Models
import sklearn.mixture
gin.external_configurable(sklearn.mixture.GaussianMixture)
gin.external_configurable(sklearn.mixture.BayesianGaussianMixture)

# Multiclass
import sklearn.multiclass
gin.external_configurable(sklearn.multiclass.OneVsRestClassifier)
gin.external_configurable(sklearn.multiclass.OneVsOneClassifier)
gin.external_configurable(sklearn.multiclass.OutputCodeClassifier)

# Multioutput
import sklearn.multioutput
gin.external_configurable(sklearn.multioutput.ClassifierChain)
gin.external_configurable(sklearn.multioutput.MultiOutputRegressor)
gin.external_configurable(sklearn.multioutput.MultiOutputClassifier)
gin.external_configurable(sklearn.multioutput.RegressorChain)

# Naive Bayes
import sklearn.naive_bayes
gin.external_configurable(sklearn.naive_bayes.BernoulliNB)
gin.external_configurable(sklearn.naive_bayes.CategoricalNB)
gin.external_configurable(sklearn.naive_bayes.ComplementNB)
gin.external_configurable(sklearn.naive_bayes.GaussianNB)
gin.external_configurable(sklearn.naive_bayes.MultinomialNB)

# Nearest Neighbors
import sklearn.neighbors
gin.external_configurable(sklearn.neighbors.BallTree)
gin.external_configurable(sklearn.neighbors.KDTree)
gin.external_configurable(sklearn.neighbors.KNeighborsClassifier)
gin.external_configurable(sklearn.neighbors.KNeighborsRegressor)
gin.external_configurable(sklearn.neighbors.LocalOutlierFactor)
gin.external_configurable(sklearn.neighbors.RadiusNeighborsClassifier)
gin.external_configurable(sklearn.neighbors.RadiusNeighborsRegressor)
gin.external_configurable(sklearn.neighbors.NearestCentroid)
gin.external_configurable(sklearn.neighbors.NearestNeighbors)
gin.external_configurable(sklearn.neighbors.NeighborhoodComponentsAnalysis)

# Pipeline
import sklearn.pipeline
gin.external_configurable(sklearn.pipeline.FeatureUnion)
gin.external_configurable(sklearn.pipeline.Pipeline)

# Preprocessing
import sklearn.preprocessing
gin.external_configurable(sklearn.preprocessing.Binarizer)
gin.external_configurable(sklearn.preprocessing.FunctionTransformer)
gin.external_configurable(sklearn.preprocessing.KBinsDiscretizer)
gin.external_configurable(sklearn.preprocessing.KernelCenterer)
gin.external_configurable(sklearn.preprocessing.LabelBinarizer)
gin.external_configurable(sklearn.preprocessing.LabelEncoder)
gin.external_configurable(sklearn.preprocessing.MultiLabelBinarizer)
gin.external_configurable(sklearn.preprocessing.MaxAbsScaler)
gin.external_configurable(sklearn.preprocessing.MinMaxScaler)
gin.external_configurable(sklearn.preprocessing.Normalizer)
gin.external_configurable(sklearn.preprocessing.OneHotEncoder)
gin.external_configurable(sklearn.preprocessing.OrdinalEncoder)
gin.external_configurable(sklearn.preprocessing.PolynomialFeatures)
gin.external_configurable(sklearn.preprocessing.PowerTransformer)
gin.external_configurable(sklearn.preprocessing.QuantileTransformer)
gin.external_configurable(sklearn.preprocessing.RobustScaler)
gin.external_configurable(sklearn.preprocessing.StandardScaler)

# Random Projection
import sklearn.random_projection
gin.external_configurable(sklearn.random_projection.GaussianRandomProjection)
gin.external_configurable(sklearn.random_projection.SparseRandomProjection)

# Support Vector Machines
import sklearn.svm
gin.external_configurable(sklearn.svm.LinearSVC)
gin.external_configurable(sklearn.svm.LinearSVR)
gin.external_configurable(sklearn.svm.NuSVC)
gin.external_configurable(sklearn.svm.NuSVR)
gin.external_configurable(sklearn.svm.OneClassSVM)
gin.external_configurable(sklearn.svm.SVC)
gin.external_configurable(sklearn.svm.SVR)

# Decision Trees
import sklearn.tree
gin.external_configurable(sklearn.tree.DecisionTreeClassifier)
gin.external_configurable(sklearn.tree.DecisionTreeRegressor)
gin.external_configurable(sklearn.tree.ExtraTreeClassifier)
gin.external_configurable(sklearn.tree.ExtraTreeRegressor)




import unittest
from itertools import cycle
from unittest.mock import MagicMock, patch

import numpy as np

from divik.cluster import _kmeans as km
from divik.cluster._kmeans import _core as cc
from divik.cluster._kmeans._core import redefine_centroids
from test.cluster.kmeans import data


class LabelingTest(unittest.TestCase):
    def setUp(self):
        self.label = km.Labeling('euclidean')
        self.centroids = np.array([
            [1, 1, 1, 1],
            [1000, 500000, -400000, -7000]
        ])

    def test_assigns_observations_to_closest_centroid(self):
        labels = self.label(data, self.centroids)
        np.testing.assert_equal(labels[:-1], 0)
        self.assertEqual(labels[-1], 1)

    def test_throws_with_no_centroids(self):
        with self.assertRaises(ValueError):
            self.label(data, np.empty((0, data.shape[1])))

    def test_throws_on_centroids_and_data_dimensionality_mismatch(self):
        smaller = data[:, :-1]
        with self.assertRaises(ValueError):
            self.label(smaller, self.centroids)


class CentroidRedefinitionTest(unittest.TestCase):
    def setUp(self):
        self.labeling = np.array([0, 0, 0, 1, 1])
        self.simple_data = np.array([
            [1, 2, 3],
            [3, 2, 1],
            [2, 2, 2],
            [10, 10, 10],
            [40, 40, 40]
        ])
        self.expected_centroids = np.array([
            [2, 2, 2],
            [25, 25, 25]
        ])

    def test_new_centroids_are_means_of_observations_in_cluster(self):
        centroids = redefine_centroids(self.simple_data, self.labeling, [0, 1])
        np.testing.assert_equal(centroids, self.expected_centroids)

    def test_throws_on_labels_and_data_length_mismatch(self):
        with self.assertRaises(ValueError):
            redefine_centroids(self.simple_data[:-1], self.labeling, [0, 1])


class KMeansTest(unittest.TestCase):
    # noinspection PyTypeChecker
    def setUp(self):
        self.mocked_initial_centroids = data[[0, -1]]
        self.mock_initialization = MagicMock(
            return_value=self.mocked_initial_centroids)
        self.mocked_labels = np.hstack([
            np.zeros((data.shape[0] - 1,)),
            np.ones((1,))
        ])
        self.mock_labeling = MagicMock(side_effect=cycle([
            self.mocked_labels,
            self.mocked_labels + 3
        ]))
        self.kmeans = km._KMeans(labeling=self.mock_labeling,
                                 initialize=self.mock_initialization,
                                 number_of_iterations=3)

    def test_only_initializes_for_no_iterations(self):
        self.kmeans.number_of_iterations = 0
        with patch.object(km._core, "redefine_centroids") as mock_redefine:
            labeling, centroids = self.kmeans(data, 2)
        self.assertEqual(0, mock_redefine.call_count)
        self.assertEqual(1, self.mock_initialization.call_count)
        self.assertEqual(1, self.mock_labeling.call_count)
        np.testing.assert_equal(self.mocked_labels, labeling)
        np.testing.assert_equal(self.mocked_initial_centroids, centroids)

    def test_initializes_once(self):
        with patch.object(km._core, "redefine_centroids"):
            self.kmeans(data, 2)
        self.assertEqual(1, self.mock_initialization.call_count)

    def test_redefines_centroids_each_iteration(self):
        with patch.object(km._core, "redefine_centroids") as mock_redefine:
            self.kmeans(data, 2)
        self.assertEqual(self.kmeans.number_of_iterations,
                         mock_redefine.call_count)

    def test_recalculates_labels_on_init_and_each_iteration(self):
        with patch.object(km._core, "redefine_centroids"):
            self.kmeans(data, 2)
        self.assertEqual(self.kmeans.number_of_iterations + 1,
                         self.mock_labeling.call_count)

    def test_returns_final_labels_and_centroids(self):
        with patch.object(km._core, "redefine_centroids") as mock_redefine:
            mock_redefine.return_value = self.mocked_initial_centroids + 3
            labeling, centroids = self.kmeans(data, 2)
        np.testing.assert_equal(centroids, self.mocked_initial_centroids + 3)
        np.testing.assert_equal(labeling, self.mocked_labels + 3)

    def test_breaks_when_reaches_stability(self):
        constant_labels = self.mocked_labels
        self.mock_labeling.side_effect = None
        self.mock_labeling.return_value = constant_labels
        with patch.object(km._core, "redefine_centroids") as mock_redefine:
            self.kmeans(data, 2)
        self.assertEqual(1, mock_redefine.call_count)

    def test_throws_for_non_2d_and_non_matrix_data(self):
        with self.assertRaises(ValueError):
            self.kmeans(data.ravel(), 2)
        with self.assertRaises(ValueError):
            # noinspection PyTypeChecker
            self.kmeans([[1, 2, 3], [4, 5, 6]], 2)


class KMeansIntegrationTest(unittest.TestCase):
    def test_segments_data_into_groups(self):
        np.random.seed(0)
        first_size = 200
        second_size = 100

        first = np.random.randn(first_size, 1)
        second = 100 + 3 * np.random.randn(second_size, 1)
        data = np.vstack([first, second])
        expected_labels = first_size * [0] + second_size * [1]

        euclidean_labeling = km.Labeling('euclidean')
        kmeans = km._KMeans(labeling=euclidean_labeling,
                            initialize=cc.ExtremeInitialization('euclidean'),
                            number_of_iterations=100)

        labels, _ = kmeans(data, 2)

        labels_are_the_same = labels == expected_labels
        labels_are_opposite = labels != expected_labels
        labels_are_valid = np.logical_or(labels_are_the_same,
                                         labels_are_opposite)
        self.assertTrue(np.all(labels_are_valid))



import logging

from divik.feature_extraction import LocallyAdjustedRbfSpectralEmbedding


def main():
    import divik._cli._utils as scr
    data, config, destination, _ = scr.initialize()
    try:
        spectral = LocallyAdjustedRbfSpectralEmbedding(**config)
        spectral.fit(data)
        spectral.save(destination)
    except Exception as ex:
        logging.error("Failed with exception.")
        logging.error(repr(ex))
        raise


if __name__ == '__main__':
    main()



import unittest
from unittest.mock import MagicMock

import numpy as np
import numpy.testing as npt

import divik.cluster._divik._backend as dv


def returns(value):
    return MagicMock(return_value=value)


def returns_many(values):
    return MagicMock(side_effect=values)


def allow_no_feature(dataset: np.ndarray):
    return np.zeros((1, dataset.shape[1])), np.inf


def allow_almost_all_features(dataset: np.ndarray):
    return np.hstack([np.zeros((1, 1)), np.ones((1, dataset.shape[1] - 1))]), np.inf


DUMMY_DATA = np.arange(100).reshape(10, 10)
SELECT_ALL = np.ones((10,), dtype=bool)


class RecursiveSelectionTest(unittest.TestCase):
    def setUp(self):
        self.selection = np.array([0, 0, 0, 1, 1, 0, 1, 0, 0, 1],
                                  dtype=bool)
        self.partition = np.array([1, 2, 2, 1], dtype=int)

    def test_selects_subset_of_elements_by_cluster_number(self):
        npt.assert_equal(
            dv._recursive_selection(self.selection, self.partition, 1),
            np.array([0, 0, 0, 1, 0, 0, 0, 0, 0, 1], dtype=bool))
        npt.assert_equal(
            dv._recursive_selection(self.selection, self.partition, 2),
            np.array([0, 0, 0, 0, 1, 0, 1, 0, 0, 0], dtype=bool))

    def test_selects_nothing_for_nonexistent_label(self):
        npt.assert_equal(
            dv._recursive_selection(self.selection, self.partition, 3),
            np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=bool))



from ._core import BaseSampler, ParallelSampler
from ._uniform_sampler import (
    UniformSampler,
    UniformPCASampler,
)
from ._stratified_sampler import StratifiedSampler

__all__ = [
    'BaseSampler',
    'ParallelSampler',
    'UniformSampler',
    'UniformPCASampler',
    'StratifiedSampler',
]



import unittest
import numpy as np
import numpy.testing as npt
from sklearn.datasets import make_blobs
from sklearn.metrics import adjusted_rand_score

from divik.cluster._divik._sklearn import DiviK
from divik._cli.dunn_divik import _fast_kmeans, _full_kmeans


fast = _fast_kmeans(distance='euclidean')
full = _full_kmeans(distance='euclidean')


class DunnDivikTest(unittest.TestCase):
    def test_predict_gives_consistent_results(self):
        X, _ = make_blobs(n_samples=100, n_features=100, centers=3,
                          random_state=42)
        model = DiviK(full, fast, distance='euclidean')
        predictions = model.fit_predict(X)
        reproduced = model.predict(X)
        npt.assert_array_equal(predictions, reproduced)

    def test_predict_works_for_single_cluster(self):
        X, _ = make_blobs(n_samples=100, n_features=100, centers=1,
                          random_state=42)
        model = DiviK(full, fast, distance='euclidean')
        predictions = model.fit_predict(X)
        reproduced = model.predict(X)
        npt.assert_array_equal(predictions, reproduced)

    def test_produces_hierarchy_for_numerous_clusters(self):
        X, _ = make_blobs(n_samples=200, n_features=100, centers=20,
                          random_state=42)
        model = DiviK(full, fast, distance='euclidean')
        model.fit(X)
        self.assertTrue(any(
            sub is not None for sub in model.result_.subregions))

    def test_predict_works_for_numerous_clusters(self):
        X, _ = make_blobs(n_samples=200, n_features=100, centers=20,
                          random_state=42)
        model = DiviK(full, fast, distance='euclidean')
        predictions = model.fit_predict(X)
        reproduced = model.predict(X)
        npt.assert_array_equal(predictions, reproduced)

    def test_works_with_pool(self):
        X, _ = make_blobs(n_samples=200, n_features=100, centers=20,
                          random_state=42)
        sequential = DiviK(full, fast, distance='euclidean')
        parallel = DiviK(full, fast, distance='euclidean', n_jobs=-1)
        expected = sequential.fit_predict(X)
        labels1 = parallel.fit_predict(X)
        labels2 = parallel.predict(X)
        npt.assert_array_equal(expected, labels1)
        npt.assert_array_equal(expected, labels2)

    def test_works_with_pool_and_sampling(self):
        X, _ = make_blobs(n_samples=1000, n_features=100, centers=5,
                          random_state=42)
        sequential = DiviK(full, fast, distance='euclidean',
                           rejection_size=5)
        parallel = DiviK(full, fast, distance='euclidean',
                         n_jobs=-1, rejection_size=5)
        expected = sequential.fit_predict(X)
        labels1 = parallel.fit_predict(X)
        labels2 = parallel.predict(X)
        npt.assert_array_equal(expected, labels1)
        npt.assert_array_equal(expected, labels2)

    def test_has_decent_performance_on_numerous_clusters(self):
        X, y = make_blobs(n_samples=200, n_features=100, centers=20,
                          random_state=42)
        model = DiviK(full, fast, distance='euclidean')
        y_pred = model.fit_predict(X)
        score = adjusted_rand_score(y, y_pred)
        self.assertGreaterEqual(score, 0.95)

    def test_transforms_to_n_clusters_dimensions(self):
        X, _ = make_blobs(n_samples=200, n_features=100, centers=20,
                          random_state=42)
        model = DiviK(full, fast, distance='euclidean')
        X_trans = model.fit_transform(X)
        self.assertEqual(X_trans.shape[1], model.n_clusters_)

    def test_is_closest_to_predicted(self):
        X, _ = make_blobs(n_samples=200, n_features=100, centers=20,
                          random_state=42)
        # This is true only with full feature's space actually, since otherwise
        # dimensionality may differ too much and influence the distance
        # comparison.
        model = DiviK(full, fast, distance='euclidean',
                      minimal_features_percentage=1.0,
                      features_percentage=1.0)
        y_pred = model.fit_predict(X)
        X_trans = model.transform(X)
        closests = [np.argmin(row) for row in X_trans]
        npt.assert_array_equal(y_pred, closests)



from copy import deepcopy

from divik._inspect.app import divik_result, xy
from divik._inspect.color import make_colormap
from divik._summary import merged_partition


_DEFAULT_CLUSTERS_FIGURE = {
    'data': [
        {
            'zsmooth': False,
            'type': 'heatmap',
            'colorscale': 'Rainbow',
            'showscale': False
        }
    ],
    'layout': {
        # 't': 40 is useful with title
        'margin': {'l': 0, 'b': 0, 'r': 0, 't': 0, 'pad': 0},
        # 'paper_bgcolor': '#7f7f7f',  # useful for debugging margins
        'autosize': True,
        'xaxis': {
            'showgrid': False,
            'showticklabels': False,
            'zeroline': False,
            'ticks': ''
        },
        'yaxis': {
            'showgrid': False,
            'showticklabels': False,
            'zeroline': False,
            'ticks': '',
            'scaleanchor': 'x',
            'scaleratio': 1
        }
    }
}


def default_clusters_figure():
    current = deepcopy(_DEFAULT_CLUSTERS_FIGURE)
    current['data'][0]['x'] = xy().T[0]
    current['data'][0]['y'] = xy().T[1].max() - xy().T[1]
    update_clusters_figure(1, [], {}, current)
    return current


def update_clusters_figure(level: int, disabled, color_overrides, current):
    partition = merged_partition(divik_result(), level)

    current['data'][0]['z'] = partition
    current['data'][0]['colorscale'] = make_colormap(partition,
                                                     disabled=disabled,
                                                     overrides=color_overrides)

    return current



import unittest

import numpy as np

import divik.feature_selection as fs


class PercentageSelectorTest(unittest.TestCase):
    def setUp(self):
        np.random.seed(42)
        self.labels = np.concatenate(
            [30 * [0] + 20 * [1] + 30 * [2] + 40 * [3]])
        self.data = np.vstack(100 * [self.labels * 10.])
        self.data += np.random.randn(*self.data.shape)
        sub = self.data[:, :-40]
        sub += 5 * np.random.randn(*sub.shape)

    def test_discards_low_abundance(self):
        p = 1 - np.sum(self.labels == 0) / self.labels.size
        selector = fs.PercentageSelector('mean', p=p).fit(self.data)
        TNR = (selector.selected_[self.labels == 0] == False).mean()
        self.assertGreaterEqual(TNR, 0.99)

    def test_discards_low_variance(self):
        p = np.sum(self.labels == 3) / self.labels.size
        selector = fs.PercentageSelector('var', p=p).fit(self.data)
        TNR = (selector.selected_[self.labels == 3] == False).mean()
        self.assertGreaterEqual(TNR, 0.99)

    def test_allows_reverse(self):
        p = 1 - np.sum(self.labels == 0) / self.labels.size
        selector = fs.PercentageSelector('mean', p=p, keep_top=False
                                         ).fit(self.data)
        TPR = selector.selected_[self.labels == 0].mean()
        self.assertGreaterEqual(TPR, 0.99)



import unittest

import numpy as np

import divik.feature_selection._outlier as fs


class HubertaTest(unittest.TestCase):
    def test_huberta(self):
        data = np.array([7.0, 6.4, 8.7, 5.8, 6.3, 6.4, 4.6, 6.1, 16.3])
        outliers = fs.huberta_outliers(data)
        self.assertEqual(outliers.sum(), 3)
        self.assertEqual(data[outliers][0], 5.8)
        self.assertEqual(data[outliers][1], 4.6)
        self.assertEqual(data[outliers][2], 16.3)


if __name__ == '__main__':
    unittest.main()



import os
import pickle
from typing import Optional

import dash
import numpy as np

from divik.core import DivikResult
from divik._cli._data_io import load_data


external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']
app = dash.Dash(__name__, external_stylesheets=external_stylesheets)
server = app.server
app.config.suppress_callback_exceptions = True

_divik_result = None
_divik_result_path = None
_xy = None


def unpickle(path: str) -> Optional[DivikResult]:
    with open(os.path.join(path, 'result.pkl'), 'rb') as result_file:
        return pickle.load(result_file)


def divik_result(path: str=None) -> Optional[DivikResult]:
    global _divik_result
    if _divik_result is None and path is not None:
        global _divik_result_path
        _divik_result_path = path
        _divik_result = unpickle(path)
    return _divik_result


def divik_result_path() -> str:
    global _divik_result_path
    return _divik_result_path


def xy(path: str=None) -> np.ndarray:
    global _xy
    if _xy is None and path is not None:
        _xy = load_data(path).astype(int)
    return _xy



import gc
from functools import partial
from typing import Optional, Union

import numpy as np
from sklearn import clone

from divik.core import Data, DivikResult
from ._report import DivikReporter


def _recursive_selection(current_selection: np.ndarray, partition: np.ndarray,
                         cluster_number: int) -> np.ndarray:
    selection = np.zeros(shape=current_selection.shape, dtype=bool)
    selection[current_selection] = partition == cluster_number
    return selection


StatSelector = 'divik.feature_selection.StatSelectorMixin'
DunnSearch = 'divik.cluster._kmeans._dunn.DunnSearch'
GAPSearch = 'divik.cluster._kmeans._gap.GAPSearch'
AutoKMeans = Union[DunnSearch, GAPSearch]


def check_stop_and_split(kmeans: AutoKMeans, fast_kmeans: GAPSearch,
                         X: Data, report: DivikReporter):
    if fast_kmeans is None:  # running GapSearch only
        report.processing(X)
        report.stop_check()
        kmeans_ = clone(kmeans).fit(X)
        if not kmeans_.fitted_ or kmeans_.n_clusters_ == 1:
            report.finished_for(X.shape[0])
            return None
    else:  # stop condition check & DunnSearch
        report.stop_check()
        fast_kmeans = clone(fast_kmeans).fit(X)
        if fast_kmeans.fitted_ and fast_kmeans.n_clusters_ == 1:
            report.finished_for(X.shape[0])
            return None
        report.processing(X)
        kmeans_ = clone(kmeans).fit(X)
    return kmeans_


# @gmrukwa: I could not find more readable solution than recursion for now.
def divik(data: Data, selection: np.ndarray,
          kmeans: AutoKMeans, fast_kmeans: GAPSearch,
          feature_selector: StatSelector,
          minimal_size: int, rejection_size: int, report: DivikReporter) \
        -> Optional[DivikResult]:
    subset = data[selection]

    if subset.shape[0] <= max(kmeans.max_clusters, minimal_size):
        report.finished_for(subset.shape[0])
        return None

    report.filter(subset)
    feature_selector = clone(feature_selector)
    filtered_data = feature_selector.fit_transform(subset)
    report.filtered(filtered_data)

    kmeans_ = check_stop_and_split(kmeans, fast_kmeans, filtered_data, report)

    if kmeans_ is None:
        return None

    partition = kmeans_.labels_
    _, counts = np.unique(partition, return_counts=True)

    if any(counts <= rejection_size):
        report.rejected(subset.shape[0])
        return None

    report.recurring(len(counts))
    recurse = partial(
        divik, data=data,
        kmeans=kmeans, fast_kmeans=fast_kmeans,
        feature_selector=feature_selector,
        minimal_size=minimal_size,
        rejection_size=rejection_size,
        report=report)
    del subset
    del filtered_data
    gc.collect()
    subregions = [
        recurse(selection=_recursive_selection(selection, partition, cluster))
        for cluster in np.unique(partition)
    ]

    report.assemble()
    return DivikResult(clustering=kmeans_, feature_selector=feature_selector,
                       merged=partition, subregions=subregions)



"""Hierarchical clustering of the data"""
from functools import partial
import logging
import os
import pickle
from typing import Callable, Dict, NamedTuple, NewType

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.cluster.hierarchy as hcl
import scipy.io as sio
from skimage.io import imsave
import divik._cli._utils as scr
import divik.core as u


LinkageMatrix = NewType('LinkageMatrix', np.ndarray)
LinkageBackend = Callable[[u.Data], LinkageMatrix]
Dendrogram = NewType('Dendrogram', Dict)
DendrogramBackend = Callable[[LinkageMatrix], Dendrogram]
SaveFigureBackend = Callable[[str], None]
Experiment = NamedTuple('Experiment', [
    ('linkage', LinkageBackend),
    ('dendrogram', DendrogramBackend),
    ('save_figure', SaveFigureBackend)
])


def flatten_linkage(linkage_matrix: LinkageMatrix) -> u.IntLabels:
    """Flatten partition of the dataset"""
    # default MATLAB behavior, seen on the dendrogram with default color
    # threshold
    cophenetic_distance_threshold = 0.7 * np.max(linkage_matrix[:, 2])
    partition = hcl.fcluster(linkage_matrix,
                             t=cophenetic_distance_threshold,
                             criterion='distance').astype(int)
    return partition


def compute_centroids(data: u.Data, partition: u.IntLabels) -> u.Data:
    """Find centroids of flat clusters"""
    return pd.DataFrame(data).groupby(partition).mean().values


def assert_configured(config, name):
    assert name in config, 'Missing "' + name + '" field in config.'


def build_experiment(config) -> Experiment:
    """Create experiment from configuration"""
    assert_configured(config, 'linkage')
    linkage_config = config['linkage']
    for item in ['method', 'metric', 'optimal_ordering']:
        assert_configured(linkage_config, item)
    linkage = partial(hcl.linkage, **linkage_config)

    assert_configured(config, 'dendrogram')
    dendrogram_config = config['dendrogram']
    for item in ['truncate_mode', 'p', 'color_threshold', 'orientation',
                 'count_sort', 'distance_sort', 'show_leaf_counts',
                 'leaf_font_size', 'show_contracted']:
        assert_configured(dendrogram_config, item)
    dendrogram = partial(hcl.dendrogram, **dendrogram_config)

    assert_configured(config, 'plot')
    plot_config = config['plot']
    for item in ['dpi', 'facecolor', 'edgecolor', 'orientation', 'transparent',
                 'frameon', 'bbox_inches', 'pad_inches']:
        assert_configured(plot_config, item)
    save_figure = partial(plt.savefig, **plot_config)

    experiment = Experiment(linkage, dendrogram, save_figure)

    return experiment


def save_linkage(fname, linkage: LinkageMatrix):
    logging.info('Saving linkage in numpy format.')
    np.save(fname('linkage.npy'), linkage)
    logging.info('Converting linkage to MATLAB format.')
    matlab_linkage = hcl.to_mlab_linkage(linkage)
    logging.info('Saving linkage in MATLAB format.')
    sio.savemat(fname('linkage.mat'), {'linkage': matlab_linkage})


def save_partition(fname, partition: u.IntLabels, xy: np.ndarray=None):
    logging.info('Saving flat partition.')
    np.save(fname('partition.npy'), partition)
    np.savetxt(fname('partition.csv'), partition, fmt='%i', delimiter=', ')
    if xy is not None:
        logging.info('Generating visulization.')
        visualization = u.visualize(partition, xy)
        imsave(fname('partition.png'), visualization)


def save_centroids(fname, centroids: np.ndarray):
    logging.info('Saving centroids.')
    np.save(fname('centroids.npy'), centroids)
    np.savetxt(fname('centroids.csv'), centroids, delimiter=', ')


def save_dendrogram(fname, save_figure: SaveFigureBackend, dendrogram: Dendrogram):
    logging.info('Pickling dendrogram data.')
    with open(fname('dendrogram.pkl'), 'wb') as file:
        pickle.dump(dendrogram, file)
    logging.info('Saving dendrogram plot as PDF.')
    save_figure(fname('dendrogram.pdf'))
    logging.info('Saving dendrogram plot as PNG.')
    save_figure(fname('dendrogram.png'))
    plt.close('all')


def main():
    """Entry point of the script"""
    data, config, destination, xy = scr.initialize()
    experiment = build_experiment(config)
    fname = partial(os.path.join, destination)
    try:
        linkage = experiment.linkage(data)
        save_linkage(fname, linkage)
        dendrogram = experiment.dendrogram(linkage)
        save_dendrogram(fname, experiment.save_figure, dendrogram)
        partition = flatten_linkage(linkage)
        save_partition(fname, partition, xy)
        centroids = compute_centroids(data, partition)
        save_centroids(fname, centroids)
    except Exception as ex:
        logging.error("Failed with exception.")
        logging.error(repr(ex))
        raise


if __name__ == '__main__':
    main()



import logging as lg
from typing import List

import numpy as np
import tqdm


def _constant_rows(matrix: np.ndarray) -> List[int]:
    is_constant = matrix.min(axis=1) == matrix.max(axis=1)
    return np.where(is_constant)[0]


class DivikReporter:
    def __init__(self, progress_reporter: tqdm.tqdm = None,
                 warn_const: bool = True):
        self.progress_reporter = progress_reporter
        self.paths_open = 1
        self.warn_const = warn_const

    def filter(self, subset):
        lg.info('Feature filtering.')
        if lg.getLogger().getEffectiveLevel() <= lg.DEBUG:
            lg.debug('Subset shape: {0}'.format(subset.shape))
            lg.debug('Has NaNs: {0}'.format(np.isnan(subset).any()))
            lg.debug('Limits: min={0}; max={1}'.format(
                subset.min(), subset.max()))
            lg.debug('Has constant rows: {0}'.format(_constant_rows(subset)))

    def filtered(self, data):
        lg.debug('Shape after filtering: {0}'.format(data.shape))
        constant = _constant_rows(data)
        if self.warn_const and any(constant):
            msg = 'After feature filtering some rows are constant: {0}. ' \
                  'This may not work with specific configurations.'
            lg.warning(msg.format(constant))

    def stop_check(self):
        lg.info('Stop condition check.')

    def finished_for(self, n_observations: int):
        self.paths_open -= 1
        lg.info('Stop condition fired for {0}. {1} paths open.'
                .format(n_observations, self.paths_open))
        if self.progress_reporter is not None:
            self.progress_reporter.update(n_observations)

    def rejected(self, n_observations: int):
        self.paths_open -= 1
        lg.info('Rejected segmentation of {0}. {1} paths open.'
                .format(n_observations, self.paths_open))
        if self.progress_reporter is not None:
            self.progress_reporter.update(n_observations)

    def processing(self, dataset: np.ndarray):
        lg.info('Processing subset with {0} observations and {1} features.'
                .format(*dataset.shape))

    def recurring(self, n_subregions):
        self.paths_open += n_subregions
        lg.info('Recurring into {0} subregions. {1} paths open.'
                .format(n_subregions, self.paths_open))

    def assemble(self):
        self.paths_open -= 1
        lg.info('Assembled. {0} paths open.'.format(self.paths_open))



from contextlib import contextmanager
from typing import Union
import uuid

from sklearn.model_selection import StratifiedShuffleSplit

from ._core import BaseSampler, ParallelSampler
from divik.core import configurable, share


_DATA = {}


@configurable
class StratifiedSampler(BaseSampler):
    """Sample the original data preserving proportions of groups

    Parameters
    -----------
    n_rows : int or float, optional (default 10000)
        Allows to limit the number of rows in the drawn samples.
        If float, should be between 0.0 and 1.0 and represent the
        proportion of the dataset to include in the sample. If
        int, represents the absolute number of rows.

    n_samples : int, optional (default None)
        Allows to limit the number of samples when iterating

    Attributes
    ----------
    X_ : array_like, shape (n_rows, n_features)
        Data to sample from

    y_ : array_like, shape (n_rows,)
        Group labels
    """
    def __init__(self, n_rows: Union[int, float] = 100,
                 n_samples: int = None):
        self.n_rows = n_rows
        self.n_samples = n_samples

    def fit(self, X, y):
        """Fit the model from data in X.

        Both inputs are preserved inside to sample from the data.

        Parameters
        ----------
        X : array-like, shape (n_rows, n_features)
            Training vector, where n_rows is the number of rows
            and n_features is the number of features.

        y: array-like, shape (n_rows,)

        Returns
        -------
        self : StratifiedSampler
            Returns the instance itself.
        """
        self.X_ = X
        self.y_ = y
        return self

    def get_sample(self, seed):
        """Return specific sample

        Sample is drawn from the set of existing rows. A proportion of
        gorups should be more-or-less the same, depending on the size
        of the sample.

        Parameters
        ----------
        seed : int
            The seed to use to draw the sample

        Returns
        -------
        sample : array_like, (*self.shape_)
            Returns the drawn sample
        """
        split = StratifiedShuffleSplit(
            n_splits=1, train_size=self.n_rows, random_state=seed)
        for idx, _ in split.split(self.X_, self.y_):
            return self.X_[idx]

    @contextmanager
    def parallel(self):
        global _DATA
        ref = str(uuid.uuid4())
        with share(self.X_) as X, share(self.y_) as y:
            _DATA[ref] = ((self.n_rows, self.n_samples), (X, y))
            try:
                yield StratifiedParallelSampler(ref)
            finally:
                del _DATA[ref]


class StratifiedParallelSampler(ParallelSampler):
    def __init__(self, ref):
        self._ref = ref

    @property
    def sampler(self):
        global _DATA
        init_args, fit_args = _DATA[self._ref]
        X, y = fit_args
        return StratifiedSampler(*init_args).fit(X.value, y.value)

    def get_sample(self, seed):
        return self.sampler.get_sample(seed)

    def initializer(self, *args):
        global _DATA
        _DATA[self._ref] = args

    @property
    def initargs(self):
        global _DATA
        return _DATA[self._ref]



from ._gin_compat import configurable, dump_gin_args, parse_gin_args
from ._parallel import get_n_jobs, maybe_pool, share
from ._seed import seed, seeded
from ._types import (
    Centroids,
    Data,
    DivikResult,
    IntLabels,
    SegmentationMethod,
)
from ._utils import build, context_if, normalize_rows, visualize


__all__ = [
    'Centroids',
    'Data',
    'DivikResult',
    'IntLabels',
    'SegmentationMethod',
    'build',
    'context_if',
    'normalize_rows',
    'visualize',
    'get_n_jobs',
    'maybe_pool',
    'share',
    'seed',
    'seeded',
    'configurable',
    'dump_gin_args',
    'parse_gin_args',
]



import numpy as np
from sklearn.base import BaseEstimator

from divik.core import configurable
from ._stat_selector_mixin import StatSelectorMixin


def _medcouple_1d(y):
    """
    Calculates the medcouple robust measure of skew.

    Parameters
    ----------
    y : array_like, 1-d
        Data to compute use in the estimator.

    Returns
    -------
    mc : float
        The medcouple statistic

    Notes
    -----
    The current algorithm requires a O(N**2) memory allocations, and so may
    not work for very large arrays (N>10000).

    .. [*] M. Huberta and E. Vandervierenb, "An adjusted boxplot for skewed
       distributions" Computational Statistics & Data Analysis, vol. 52, pp.
       5186-5201, August 2008.
    """

    # Parameter changes the algorithm to the slower for large n

    y = np.squeeze(np.asarray(y))
    if y.ndim != 1:
        raise ValueError("y must be squeezable to a 1-d array")

    y = np.sort(y)

    n = y.shape[0]
    if n % 2 == 0:
        mf = (y[n // 2 - 1] + y[n // 2]) / 2
    else:
        mf = y[(n - 1) // 2]

    z = y - mf
    lower = z[z <= 0.0]
    upper = z[z >= 0.0]
    upper = upper[:, None]
    standardization = upper - lower
    is_zero = np.logical_and(lower == 0.0, upper == 0.0)
    standardization[is_zero] = np.inf
    spread = upper + lower
    h = spread / standardization
    # GH5395
    num_ties = np.sum(lower == 0.0)
    if num_ties:
        # Replacements has -1 above the anti-diagonal, 0 on the anti-diagonal,
        # and 1 below the anti-diagonal
        replacements = np.ones((num_ties, num_ties)) - np.eye(num_ties)
        replacements -= 2 * np.triu(replacements)
        # Convert diagonal to anti-diagonal
        replacements = np.fliplr(replacements)
        # Always replace upper right block
        h[:num_ties, -num_ties:] = replacements

    return np.median(h)


def medcouple(y, axis=0):
    """
    Calculate the medcouple robust measure of skew.

    Parameters
    ----------
    y : array_like
        Data to compute use in the estimator.
    axis : {int, None}
        Axis along which the medcouple statistic is computed.  If `None`, the
        entire array is used.

    Returns
    -------
    mc : ndarray
        The medcouple statistic with the same shape as `y`, with the specified
        axis removed.

    Notes
    -----
    The current algorithm requires a O(N**2) memory allocations, and so may
    not work for very large arrays (N>10000).

    .. [*] M. Huberta and E. Vandervierenb, "An adjusted boxplot for skewed
       distributions" Computational Statistics & Data Analysis, vol. 52, pp.
       5186-5201, August 2008.
    """
    y = np.asarray(y, dtype=np.double)  # GH 4243
    if axis is None:
        return _medcouple_1d(y.ravel())

    return np.apply_along_axis(_medcouple_1d, axis, y)


def huberta_outliers(v):
    """
    M. Huberta, E.Vandervierenb (2008) An adjusted boxplot for skewed
    distributions, Computational Statistics and Data Analysis 52 (2008)
    5186–5201

    Parameters
    ----------
    v: array-like
        An array to filter outlier from.

    Returns
    -------
    Binary vector indicating all the outliers.
    """
    q1, q3 = np.quantile(v, q=[.25, .75])
    iqr = q3 - q1
    MC = medcouple(v)
    if MC >= 0:
        lower_bound = q1 - 1.5 * np.exp(-4 * MC) * iqr
        upper_bound = q3 + 1.5 * np.exp(3 * MC) * iqr
    else:
        lower_bound = q1 - 1.5 * np.exp(-3 * MC) * iqr
        upper_bound = q3 + 1.5 * np.exp(4 * MC) * iqr
    return np.logical_or(v < lower_bound, v > upper_bound)


# noinspection PyAttributeOutsideInit
@configurable
class OutlierSelector(BaseEstimator, StatSelectorMixin):
    """Feature selector that removes outlier features w.r.t. mean or variance

    Huberta's outlier detection is applied to the features' characteristics
    and the outlying features are removed.

    This feature selection algorithm looks only at the features (X), not the
    desired outputs (y), and can thus be used for unsupervised learning.

    Parameters
    ----------
    stat: {'mean', 'var'}
        Kind of statistic to be computed out of the feature.

    use_log: bool, optional, default: False
        Whether to use the logarithm of feature characteristic instead of the
        characteristic itself. This may improve feature filtering performance,
        depending on the distribution of features, however all the
        characteristics (mean, variance) have to be positive for that -
        filtering will fail otherwise. This is useful for specific cases in
        biology where the distribution of data may actually require this option
        for any efficient filtering.

    keep_outliers: bool, optional, default: False
        When True, keeps outliers instead of inlier features.

    Attributes
    ----------
    vals_: array, shape (n_features,)
        Computed characteristic of each feature.

    selected_: array, shape (n_features,)
        Vector of binary selections of the informative features.
    """
    def __init__(self, stat: str, use_log: bool = False,
                 keep_outliers: bool = False):
        self.stat = stat
        self.use_log = use_log
        self.keep_outliers = keep_outliers

    def fit(self, X, y=None):
        """Learn data-driven feature thresholds from X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            Sample vectors from which to compute feature characteristic.

        y : any
            Ignored. This parameter exists only for compatibility with
            sklearn.pipeline.Pipeline.

        Returns
        -------
        self
        """
        self.vals_ = self._to_characteristics(X)
        outliers = huberta_outliers(self.vals_)
        if self.keep_outliers:
            self.selected_ = outliers
        else:
            self.selected_ = outliers == False
        return self



import json

from dash.dependencies import Output, Input, State

from divik._inspect import figure as fig
from divik._inspect.app import app
from divik._inspect.layout import Fields


@app.callback(
    Output(Fields.CLUSTERS_GRAPH, 'figure'),
    [
        Input(Fields.LEVEL, 'value'),
        Input(Fields.DISABLED_CLUSTERS_STORAGE, 'children'),
        Input(Fields.COLOR_OVERRIDES_STORAGE, 'children'),
    ],
    [State(Fields.CLUSTERS_GRAPH, 'figure')]
)
def update_visualization(depth, disabled_state, overrides, figure):
    disabled = json.loads(disabled_state)['excluded'] if disabled_state else []
    overrides = json.loads(overrides).get(str(depth), {}) if overrides else {}
    return fig.update_clusters_figure(depth, disabled, overrides, figure)



import inspect
import six

import gin


# TODO: Delete this file when https://github.com/google/gin-config/pull/48 gets merged
def _decorate_fn_or_cls(decorator, fn_or_cls, subclass=False):
  """Decorate a function or class with the given decorator.

  When `fn_or_cls` is a function, applies `decorator` to the function and
  returns the (decorated) result.

  When `fn_or_cls` is a class and the `subclass` parameter is `False`, this will
  replace `fn_or_cls.__init__` with the result of applying `decorator` to it.

  When `fn_or_cls` is a class and `subclass` is `True`, this will subclass the
  class, but with `__init__` defined to be the result of applying `decorator` to
  `fn_or_cls.__init__`. The decorated class has metadata (docstring, name, and
  module information) copied over from `fn_or_cls`. The goal is to provide a
  decorated class the behaves as much like the original as possible, without
  modifying it (for example, inspection operations using `isinstance` or
  `issubclass` should behave the same way as on the original class).

  Args:
    decorator: The decorator to use.
    fn_or_cls: The function or class to decorate.
    subclass: Whether to decorate classes by subclassing. This argument is
      ignored if `fn_or_cls` is not a class.

  Returns:
    The decorated function or class.
  """
  if not inspect.isclass(fn_or_cls):
    return decorator(gin.config._ensure_wrappability(fn_or_cls))

  construction_fn = gin.config._find_class_construction_fn(fn_or_cls)

  if subclass:
    class DecoratedClass(fn_or_cls):
      __doc__ = fn_or_cls.__doc__
      __module__ = fn_or_cls.__module__
    DecoratedClass.__name__ = fn_or_cls.__name__
    if six.PY3:
      DecoratedClass.__qualname__ = fn_or_cls.__qualname__
    cls = DecoratedClass
  else:
    cls = fn_or_cls

  decorated_fn = decorator(gin.config._ensure_wrappability(construction_fn))
  if construction_fn.__name__ == '__new__':
    decorated_fn = staticmethod(decorated_fn)
  elif subclass:
    # this ensures picklability, as we produce the original class instance,
    # the implicit inheritance is skipped
    def hijacking_new(klass, *args, **kwargs):
      if klass is DecoratedClass:
        klass = fn_or_cls
      obj = fn_or_cls.__new__(klass, *args, **kwargs)
      decorated_fn(obj, *args, **kwargs)
      return obj
    setattr(cls, '__new__', hijacking_new)
  setattr(cls, construction_fn.__name__, decorated_fn)
  return cls


gin.config._decorate_fn_or_cls = _decorate_fn_or_cls



"""Summarizing functions for DiviK result.

_summary.py

Copyright 2019 Grzegorz Mrukwa

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
from typing import Dict, List, Optional, Tuple

import numpy as np

import divik.core as u


def depth(tree, children_collection_name='subregions'):
    """Get tree depth."""
    if tree is None:
        return 1
    return max(depth(subtree) for subtree
               in getattr(tree, children_collection_name)) + 1


def total_number_of_clusters(tree) -> int:
    """Get the number of leaves in the tree."""
    if tree is None:
        return 1
    return sum(total_number_of_clusters(subtree)
               for subtree in tree.subregions)


def merged_partition(tree: u.DivikResult, levels_limit: int = np.inf,
                     return_paths: bool = False) -> u.IntLabels:
    """Compute merged segmentation labels."""
    assert tree is not None, 'Nothing was segmented.'
    merged, paths = _merged_partition(
        tree.clustering.labels_, tree.subregions, levels_limit)
    if return_paths:
        return merged, paths
    return merged


def _merged_partition(partition: u.IntLabels,
                      subregions: List[Optional[u.DivikResult]],
                      levels_limit: int = np.inf) \
        -> Tuple[u.IntLabels, Dict[int, Tuple[int]]]:
    """Compute merged segmentation labels."""
    result = partition * 0 - 1
    known_clusters = 0
    paths = {}
    for cluster_number, subregion in enumerate(subregions):
        current_cluster = partition == cluster_number
        if subregion is None or levels_limit <= 1:
            result[current_cluster] = known_clusters
            paths[known_clusters] = (cluster_number,)
            known_clusters += 1
        else:
            local_partition, down_paths = _merged_partition(
                subregion.clustering.labels_, subregion.subregions,
                levels_limit-1)
            result[current_cluster] = local_partition + known_clusters
            for cluster in np.unique(local_partition):
                paths[cluster + known_clusters] = (
                    cluster_number, *down_paths[cluster])
            known_clusters += int(np.max(local_partition)) + 1
    return result, paths


def _update_graph(tree, size: int, graph: 'networkx.Graph' = None,
                  parent=None):
    tree_node = len(graph)
    graph.add_node(tree_node, size=size)
    if parent is not None:
        graph.add_edge(parent, tree_node)
    if tree is None:
        return
    for idx, subtree in enumerate(tree.subregions):
        _update_graph(subtree,
                      size=np.sum(tree.clustering.labels_ == idx),
                      graph=graph,
                      parent=tree_node)


def _as_graph(tree) -> 'networkx.DiGraph':
    import networkx as nx
    graph = nx.DiGraph()
    _update_graph(tree, size=tree.partition.size, graph=graph, parent=None)
    return graph


def _make_labels(graph):
    import networkx as nx
    attributes = nx.get_node_attributes(graph, 'size')
    return {
        idx: "{0} (size={1})".format(idx, attributes[idx])
        for idx in graph
    }


def _make_sizes(graph):
    import networkx as nx
    values = np.array(list(nx.get_node_attributes(graph, 'size').values()))
    return list(np.sqrt(values))


def scale_plot_size(factor=1.5):
    """Scale plot size in jupyter notebook."""
    import matplotlib as mpl
    default_dpi = mpl.rcParamsDefault['figure.dpi']
    mpl.rcParams['figure.dpi'] = default_dpi * factor


def _make_positions(graph):
    import networkx as nx
    intermediate = nx.spectral_layout(graph)
    return nx.spring_layout(graph, pos=intermediate, iterations=20)


def plot(tree, with_size=False):
    """Plot visualization of splits."""
    graph = _as_graph(tree)
    arguments = {
        'G': graph,
        'pos': _make_positions(graph),
        'node_size': _make_sizes(graph),
        'font_size': 3,
        'width': .1  # lines width
    }
    if with_size:
        arguments['labels'] = _make_labels(graph)
    import networkx as nx
    nx.draw_networkx(**arguments)
    import matplotlib.pyplot as plt
    plt.axis('off')


def reject_split(tree: Optional[u.DivikResult],
                 rejection_size: int = 0) \
        -> Optional[u.DivikResult]:
    """Re-apply rejection condition on known result tree."""
    if tree is None:
        logging.debug("Rejecting empty.")
        return None
    counts = np.unique(tree.clustering.labels_, return_counts=True)[1]
    if any(counts <= rejection_size):
        logging.debug("Rejecting by condition.")
        return None
    allowed_subregions = [
        reject_split(subregion, rejection_size) for subregion in tree.subregions
    ]
    merged, _ = _merged_partition(tree.clustering.labels_, allowed_subregions)
    logging.debug("Returning pruned tree.")
    return u.DivikResult(
        clustering=tree.clustering,
        feature_selector=tree.feature_selector,
        merged=merged,
        subregions=allowed_subregions
    )



from abc import ABCMeta, abstractmethod
from functools import partial
import logging
from typing import List, NamedTuple, Union

import numpy as np
import scipy.spatial.distance as dist
from sklearn.linear_model import LinearRegression

from divik.core import Centroids, Data


class Initialization(object, metaclass=ABCMeta):
    """Initializes k-means algorithm"""
    @abstractmethod
    def __call__(self, data: Data, number_of_centroids: int) -> Centroids:
        """Generate initial centroids for k-means algorithm

        @param data: 2D matrix with observations in rows, features in columns
        @param number_of_centroids: number of centroids to be generated
        @return: centroids, in rows
        """
        raise NotImplementedError(self.__class__.__name__
                                  + " must implement __call__.")


def _find_residuals(data: Data, sample_weight=None) -> np.ndarray:
    features = data.T
    assumed_ys = features[0]
    modelled_xs = np.hstack([np.ones((data.shape[0], 1)),
                            features[1:].T])
    lr = LinearRegression().fit(modelled_xs, assumed_ys,
                                sample_weight=sample_weight)
    residuals = np.abs(lr.predict(modelled_xs) - assumed_ys)
    return residuals


def _validate(data: Data, number_of_centroids: int):
    if number_of_centroids > data.shape[0]:
        msg = f"Number of centroids ({number_of_centroids}) greater than " + \
            f"number of observations ({data.shape[0]})"
        logging.error(msg)
        raise ValueError(msg)
    if number_of_centroids < 1:
        msg = f'number_of_centroids({number_of_centroids}) < 1'
        logging.error(msg)
        raise ValueError(msg)


class ExtremeInitialization(Initialization):
    """Initializes k-means by picking extreme points"""
    def __init__(self, distance: str):
        self.distance = distance

    def __call__(self, data: Data, number_of_centroids: int) -> Centroids:
        """Generate initial centroids for k-means algorithm

        @param data: 2D matrix with observations in rows, features in columns
        @param number_of_centroids: number of centroids to be generated
        @return: centroids, in rows
        """
        _validate(data, number_of_centroids)
        residuals = _find_residuals(data)
        centroids = np.nan * np.zeros((number_of_centroids, data.shape[1]))
        centroids[0] = data[np.argmax(residuals)]

        distances = np.inf * np.ones((data.shape[0], ))
        for i in range(1, number_of_centroids):
            current_distance = dist.cdist(
                data, centroids[np.newaxis, i - 1], self.distance)
            distances[:] = np.minimum(current_distance.ravel(), distances)
            centroids[i] = data[np.argmax(distances)]

        return centroids


class PercentileInitialization(Initialization):
    def __init__(self, distance: str, percentile: float=99.):
        assert 0 <= percentile <= 100, percentile
        self.distance = distance
        self.percentile = percentile

    def _get_percentile_element(self, values: np.ndarray) -> int:
        value = np.percentile(values, q=self.percentile,
                              interpolation='nearest')
        assert values.size > 0
        assert not np.isnan(values).any()
        matches = values == value
        assert np.any(matches), (value, values)
        return int(np.flatnonzero(matches)[0])

    def __call__(self, data: Data, number_of_centroids: int) -> Centroids:
        _validate(data, number_of_centroids)
        residuals = _find_residuals(data)
        selected = self._get_percentile_element(residuals)
        centroids = np.nan * np.zeros((number_of_centroids, data.shape[1]))
        centroids[0] = data[selected]
        assert not np.any(np.isnan(centroids[0]))

        distances = np.inf * np.ones((data.shape[0],))
        for i in range(1, number_of_centroids):
            assert not np.any(np.isnan(centroids[np.newaxis, i - 1]))
            current_distance = dist.cdist(
                data, centroids[np.newaxis, i - 1], self.distance)
            nans = np.isnan(current_distance)
            if np.any(nans):
                locations_of_nans = np.array(list(zip(*np.nonzero(nans))))
                msg = 'Distances between points cannot be NaN. This ' + \
                    'indicates that your data is probably corrupted and ' + \
                    'analysis cannot be continued in this setting. Amount' + \
                    f' of NaNs: {nans.sum()}. At positions described by ' + \
                    f'[spot, centroid]: {locations_of_nans}'
                logging.error(msg)
                raise ValueError(msg)
            distances[:] = np.minimum(current_distance.ravel(), distances)
            selected = self._get_percentile_element(distances)
            centroids[i] = data[selected]

        return centroids


class Leaf(NamedTuple):
    centroid: np.ndarray
    count: int = 0

KDTree = Union['Node', Leaf]
        
class Node(NamedTuple):
    left: KDTree = None
    right: KDTree = None


def make_tree(X, leaf_size: int, _feature_idx: int = 0, selector=None) \
        -> KDTree:
    """Make KDTree out of the data

    Construct a KDTree out of data using mean as a pivoting element.
    Each split makes two segments. The result doesn't contain the original
    data, just centroids in each box and count of items.

    Parameters
    ==========
    X : array_like, (n_samples, n_features)
        Set of observations to divide into boxes
        
    leaf_size : int
        Desired leaf size. It should more than `leaf_size` and
        will be up to `2 * leaf_size`
    
    Returns
    =======
    tree : KDTree
        Lightweight KD-Tree over the data
    """
    if selector is None:
        selector = np.ones((X.shape[0],), dtype=bool)
    if selector.sum() < 2 * leaf_size:
        centroid = X[selector, :].mean(axis=0, keepdims=True)
        return Leaf(centroid, X.shape[0])
    feature = X[selector, _feature_idx]
    thr = np.mean(feature)
    left_idx = selector.copy()
    left_idx[selector] = feature < thr
    right_idx = selector.copy()
    right_idx[selector] = np.logical_not(left_idx[selector])
    next_feature = (_feature_idx + 1) % X.shape[1]
    return Node(
        left=make_tree(X, leaf_size=leaf_size,
            _feature_idx=next_feature, selector=left_idx),
        right=make_tree(X, leaf_size=leaf_size,
            _feature_idx=next_feature, selector=right_idx),
    )


def get_leaves(tree: KDTree) -> List[Leaf]:
    """Extract leaves of the KDTree
    
    Parameters
    ==========
    tree : KDTree
        KDTree constructed on the data
        
    Returns
    =======
    leaves : list of Leaf
        All the leaves from the full depth of the tree
    """
    if isinstance(tree, Leaf):
        return [tree]
    return get_leaves(tree.left) + get_leaves(tree.right)


class KDTreeInitialization(Initialization):
    """Initializes k-means by picking extreme KDTree box"""
    def __init__(self, distance: str, leaf_size: Union[int, float] = 0.01):
        self.distance = distance
        self.leaf_size = leaf_size

    def __call__(self, data: Data, number_of_centroids: int) -> Centroids:
        """Generate initial centroids for k-means algorithm"""
        _validate(data, number_of_centroids)
        leaf_size = self.leaf_size
        if isinstance(leaf_size, float):
            if 0 <= leaf_size <= 1:
                leaf_size = max(int(leaf_size * data.shape[0]), 1)
            else:
                logging.error('leaf_size must be between 0 and 1 when float')
                raise ValueError('leaf_size must be between 0 and 1 when float')
        tree = make_tree(data, leaf_size=leaf_size)
        leaves = get_leaves(tree)
        box_centroids = np.vstack([l.centroid for l in leaves])
        box_weights = np.array([l.count for l in leaves])

        residuals = _find_residuals(box_centroids, box_weights)
        centroids = np.nan * np.zeros((number_of_centroids, data.shape[1]))
        centroids[0] = box_centroids[np.argmax(residuals)]

        distances = np.inf * np.ones((box_centroids.shape[0], ))
        for i in range(1, number_of_centroids):
            current_distance = dist.cdist(
                box_centroids, centroids[np.newaxis, i - 1], self.distance)
            distances[:] = np.minimum(current_distance.ravel(), distances)
            centroids[i] = box_centroids[np.argmax(distances)]

        return centroids


class KDTreePercentileInitialization(Initialization):
    """Initializes k-means by picking extreme KDTree box"""
    def __init__(self, distance: str, leaf_size: Union[int, float] = 0.01,
                 percentile: float=99.):
        assert 0 <= percentile <= 100, percentile
        self.distance = distance
        self.leaf_size = leaf_size
        self.percentile = percentile

    def _get_percentile_idx(self, distances, weights) -> int:
        idx = np.argsort(distances)
        over_percentile = np.cumsum(weights[idx]) >= self.percentile
        first_over = np.flatnonzero(over_percentile)[0]
        return idx[first_over]

    def __call__(self, data: Data, number_of_centroids: int) -> Centroids:
        """Generate initial centroids for k-means algorithm"""
        _validate(data, number_of_centroids)
        leaf_size = self.leaf_size
        if isinstance(leaf_size, float):
            if 0 <= leaf_size <= 1:
                leaf_size = max(int(leaf_size * data.shape[0]), 1)
            else:
                logging.error('leaf_size must be between 0 and 1 when float')
                raise ValueError('leaf_size must be between 0 and 1 when float')
        tree = make_tree(data, leaf_size=leaf_size)
        leaves = get_leaves(tree)
        box_centroids = np.vstack([l.centroid for l in leaves])
        box_weights = np.array([l.count for l in leaves])
        normalized_weights = 100 * box_weights / np.sum(box_weights)

        residuals = _find_residuals(box_centroids, box_weights)
        centroids = np.nan * np.zeros((number_of_centroids, data.shape[1]))
        idx = self._get_percentile_idx(residuals, normalized_weights)
        centroids[0] = box_centroids[idx]

        distances = np.inf * np.ones((box_centroids.shape[0], ))
        for i in range(1, number_of_centroids):
            current_distance = dist.cdist(
                box_centroids, centroids[np.newaxis, i - 1], self.distance)
            distances[:] = np.minimum(current_distance.ravel(), distances)
            idx = self._get_percentile_idx(distances, normalized_weights)
            centroids[i] = box_centroids[idx]

        return centroids



from dash.dependencies import Output, Input, State

from divik._inspect import persistence as per
from divik._inspect.app import app
from divik._inspect.layout import Fields


@app.callback(
    Output(Fields.SAVED_PROFILES, 'options'),
    [Input(Fields.SAVE_PROFILE, 'n_clicks')],
    [
        State(Fields.NEW_PROFILE_NAME, 'value'),
        State(Fields.DISABLED_CLUSTERS_STORAGE, 'children'),
        State(Fields.COLOR_OVERRIDES_STORAGE, 'children'),
        State(Fields.LEVEL, 'value'),
    ]
)
def save_profile(_, name, disabled, color, level):
    if name:
        per.preserve_by_name(name, disabled, color, level)
    return per.find_preserved()


@app.callback(
    Output(Fields.LEVEL, 'value'),
    [Input(Fields.LOAD_PROFILE, 'n_clicks')],
    [
        State(Fields.SAVED_PROFILES, 'value'),
        State(Fields.LEVEL, 'value')
    ]
)
def load_level_from_profile(_, name, self):
    if not name:
        return self
    return per.restore_level(name)



import unittest

from collections import namedtuple

import numpy as np
import numpy.testing as npt

import divik._summary as sm
import divik.core as u


DummyClustering = namedtuple(
    'DummyClustering', ('cluster_centers_', 'best_score_', 'labels_'))


DUMMY_RESULT = u.DivikResult(
    clustering=DummyClustering(
        cluster_centers_=np.zeros((3, 1)),
        best_score_=3.,
        labels_=np.array([0] * 10 + [1] * 5 + [2] * 10, dtype=int)
    ),
    feature_selector=None,
    merged=np.array([0] * 10 + [1] * 15, dtype=int),
    subregions=[
        None,
        u.DivikResult(
            clustering=DummyClustering(
                cluster_centers_=np.zeros((2, 1)),
                best_score_=2.,
                labels_=np.array([0] + [1] * 4, dtype=int)
            ),
            feature_selector=None,
            merged=np.array([0] + [1] * 4, dtype=int),
            subregions=[None, None]
        ),
        u.DivikResult(
            clustering=DummyClustering(
                cluster_centers_=np.zeros((3, 1)),
                best_score_=2.,
                labels_=np.array([0] * 2 + [1] * 3 + [2] * 5, dtype=int)
            ),
            feature_selector=None,
            merged=np.array([0] * 2 + [1] * 3 + [2] * 5, dtype=int),
            subregions=[None, None, None]
        )
    ]
)


class DepthTest(unittest.TestCase):
    def test_resolves_tree_depth(self):
        self.assertEqual(sm.depth(DUMMY_RESULT), 3)


class MergeTest(unittest.TestCase):
    def test_merges_disjoint_regions(self):
        partition = sm.merged_partition(DUMMY_RESULT)
        regions, counts = np.unique(partition, return_counts=True)
        npt.assert_equal([10, 1, 4, 2, 3, 5], counts)
        npt.assert_equal(np.arange(6), regions)

    def test_returns_paths_to_partitions(self):
        partition, paths = sm.merged_partition(DUMMY_RESULT, return_paths=True)
        self.assertEqual(paths[0], (0,))
        self.assertEqual(paths[1], (1, 0))
        self.assertEqual(paths[4], (2, 1))
        self.assertNotIn(6, paths)


class RejectionTest(unittest.TestCase):
    def test_without_rejection_updates_merged_and_nothing_else(self):
        filtered = sm.reject_split(DUMMY_RESULT, 0)
        self.assertEqual(filtered.clustering.best_score_,
                         DUMMY_RESULT.clustering.best_score_)
        self.assertEqual(sm.depth(filtered), sm.depth(DUMMY_RESULT))
        npt.assert_equal(filtered.merged, sm.merged_partition(DUMMY_RESULT))

    def test_rejects_splits(self):
        filtered = sm.reject_split(DUMMY_RESULT, 2)
        self.assertIsNone(filtered.subregions[1])



#!/usr/bin/env python
import json
import logging
import os
import pickle
import typing
import numpy as np
import pandas as pd
import skimage.io as sio

from divik.core import build
from divik._cli._data_io import DIVIK_RESULT_FNAME
from divik.cluster import DiviK, GAPSearch, KMeans
import divik._summary as _smr
import divik._cli._utils as sc
import divik.core as u


def _make_summary(result: typing.Optional[u.DivikResult]):
    n_clusters = _smr.total_number_of_clusters(result)
    if result is not None:
        mean_cluster_size = result.clustering.labels_.shape[0] / float(n_clusters)
    else:
        mean_cluster_size = -1
    return {
        "depth": _smr.depth(result),
        "number_of_clusters": n_clusters,
        "mean_cluster_size": mean_cluster_size
    }


def make_merged(result: typing.Optional[u.DivikResult]) -> np.ndarray:
    depth = _smr.depth(result)
    return np.hstack([
        _smr.merged_partition(result, limit + 1).reshape(-1, 1)
        for limit in range(depth)
    ])


def save_merged(destination: str, merged: np.ndarray, xy: np.ndarray=None):
    np.savetxt(os.path.join(destination, 'partitions.csv'),
               merged, delimiter=', ', fmt='%i')
    np.save(os.path.join(destination, 'partitions.npy'), merged)
    if xy is not None:
        for level in range(merged.shape[1]):
            np.save(
                os.path.join(destination, 'partition-{0}.npy'.format(level)),
                merged[:, level]
            )
            visualization = u.visualize(merged[:, level], xy=xy)
            image_name = os.path.join(destination, 'partition-{0}.png'.format(level))
            sio.imsave(image_name, visualization)
    final_partition = merged[:, -1]
    np.save(os.path.join(destination, 'final_partition.npy'), final_partition)
    np.savetxt(os.path.join(destination, 'final_partition.csv'), final_partition,
               delimiter=', ', fmt='%i')


def save(data: u.Data, divik: DiviK, destination: str, xy: np.ndarray=None):
    logging.info("Saving result.")
    logging.info("Saving pickle.")
    with open(os.path.join(destination, DIVIK_RESULT_FNAME), 'wb') as pkl:
        pickle.dump(divik.result_, pkl)
    with open(os.path.join(destination, 'model.pkl'), 'wb') as pkl:
        pickle.dump(divik, pkl)
    logging.info("Saving JSON summary.")
    with open(os.path.join(destination, 'summary.json'), 'w') as smr:
        json.dump(_make_summary(divik.result_), smr)
    if divik.result_ is not None:
        logging.info("Saving partitions.")
        merged = make_merged(divik.result_)
        assert merged.shape[0] == divik.result_.clustering.labels_.size
        save_merged(destination, merged, xy)
        logging.info("Saving centroids.")
        centroids = pd.DataFrame(data).groupby(merged[:, -1]).mean().values
        np.save(os.path.join(destination, 'centroids.npy'), centroids)
        np.savetxt(os.path.join(destination, 'centroids.csv'), centroids,
                   delimiter=', ')
    else:
        logging.info("Skipping partition save. Cause: result is None")


def _full_kmeans(**config):
    distance = config.get('distance', 'correlation')
    normalize_rows = config.get('normalize_rows', None)
    single_kmeans = KMeans(
        n_clusters=2,
        distance=distance,
        init='percentile',
        percentile=config.get('distance_percentile', 99.0),
        max_iter=config.get('max_iter', 100),
        normalize_rows=normalize_rows,
    )
    kmeans = GAPSearch(
        single_kmeans,
        max_clusters=config.get('k_max', 50),
        n_jobs=config.get('n_jobs', None),
        seed=config.get('random_seed', 0),
        n_trials=config.get('gap_trials', 10),
        sample_size=config.get('sample_size', 10000),
        verbose=config.get('verbose', False),
    )
    return kmeans


def main():
    data, config, destination, xy = sc.initialize()
    logging.info('Workspace initialized.')
    logging.info('Scenario configuration: {0}'.format(config))
    fast = None
    full = _full_kmeans(**config)
    divik_config = {'kmeans': full, 'fast_kmeans': fast, **config}
    divik = build(DiviK, **divik_config)
    logging.info("Launching experiment.")
    try:
        divik.fit(data)
    except Exception as ex:
        logging.error("Failed with exception.")
        logging.error(repr(ex))
        raise
    save(data, divik, destination, xy)


if __name__ == '__main__':
    main()



import argparse as _agp
import json
import logging
import os
import sys
import time
from typing import Dict, Tuple

import numpy as np
import pandas as pd
import tqdm

import divik.core as u
from divik import __version__
from divik._cli._data_io import load_data


def parse_args():
    parser = _agp.ArgumentParser()
    parser.add_argument('--source', '-s',
                        help='Source file with data. Should contain'
                             ' observations in rows and features in columns.',
                        action='store', dest='source', required=True)
    parser.add_argument('--destination', '-d',
                        help='Destination directory into which results will be'
                             ' saved.',
                        action='store', dest='destination', required=True)
    parser.add_argument('--config', '-c',
                        help='Configuration file for the experiment.',
                        action='store', dest='config', required=True)
    parser.add_argument('--xy',
                        help='File with spatial coordinates X and Y. Should '
                             'contain X in first column, Y in second. The '
                             'number of rows should be equal to the data.',
                        action='store', dest='xy', required=False, default=None)
    parser.add_argument('--omit-datetime', action='store_true',
                        help='If defines, omit the datetime subdirectory of '
                        'result.')
    parser.add_argument('--verbose', '-v', action='store_true')
    return parser.parse_args()


def prepare_destination(destination: str, omit_datetime: bool = False,
                        exist_ok: bool = False) -> str:
    if not omit_datetime:
        datetime = time.strftime("%Y%m%d-%H%M%S")
        destination = os.path.join(destination, datetime)
    os.makedirs(destination, exist_ok=exist_ok)
    return destination


_PRECISE_FORMAT = '%(asctime)s [%(levelname)s] %(filename)40s:%(lineno)3s' \
    + ' - %(funcName)40s\t%(message)s'
_INFO_FORMAT = '%(asctime)s [%(levelname)s]\t%(message)s'


def _file_handler(destination: str):
    log_destination = os.path.join(destination, 'logs.txt')
    handler = logging.FileHandler(filename=log_destination, mode='a')
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(logging.Formatter(_PRECISE_FORMAT))
    return handler


def _stream_handler(verbose: bool):
    handler = logging.StreamHandler(tqdm.tqdm)
    if verbose:
        handler.setLevel(logging.INFO)
        handler.setFormatter(logging.Formatter(_INFO_FORMAT))
    else:
        handler.setLevel(logging.CRITICAL)
        handler.setFormatter(logging.Formatter(_INFO_FORMAT))
    return handler


def setup_logger(destination: str, verbose: bool = False):
    stream_handler = _stream_handler(verbose)
    file_handler = _file_handler(destination)
    handlers = [
        stream_handler,
        file_handler,
    ]
    del logging.root.handlers[:]
    logging.basicConfig(level=logging.DEBUG, handlers=handlers)
    version_notice = "Using " + sys.argv[0] + \
                     " (divik, version " + __version__ + ")"
    logging.info(version_notice)


def load_config(config_path, destination: str):
    logging.info("Reading configuration file: " + config_path)
    with open(config_path) as configfile:
        config = json.load(configfile)
    logging.info('Backing up configuration.')
    with open(os.path.join(destination, 'config.json'), 'w') as configfile:
        json.dump(config, configfile)
    return config


Config = Dict
DestinationPath = str
Coordinates = np.ndarray


def try_load_data(path):
    try:
        data = load_data(path)
        logging.debug('Data loaded successfully.')
    except Exception as ex:
        logging.error("Data loading failed with an exception.")
        logging.error(repr(ex))
        raise
    return data


def try_load_xy(path):
    if path is not None:
        try:
            xy = load_data(path).astype(int)
            logging.debug('Coordinates loaded successfully.')
        except Exception as ex:
            logging.error('Coordinates loading failed with an exception.')
            logging.error(repr(ex))
            raise
    else:
        xy = None
    return xy


def initialize() -> Tuple[u.Data, Config, DestinationPath, Coordinates]:
    arguments = parse_args()
    destination = prepare_destination(arguments.destination,
                                      arguments.omit_datetime)
    setup_logger(destination, arguments.verbose)
    config = load_config(arguments.config, destination)
    data = try_load_data(arguments.source)
    xy = try_load_xy(arguments.xy)
    return data, config, destination, xy


def save_csv(array: np.ndarray, fname: str):
    pd.DataFrame(array).to_csv(fname, header=False, index=False)



"""Clustering methods"""
from ._kmeans import (
    DunnSearch,
    GAPSearch,
    KMeans,
)
from ._divik import (
    DiviK,
)

__all__ = [
    'DiviK',
    'DunnSearch',
    'GAPSearch',
    'KMeans',
]



import argparse as agp

from divik._cli._data_io import as_divik_result_path
from divik._inspect.app import app, divik_result, xy
# noinspection PyUnresolvedReferences
import divik._inspect.callback
from divik._inspect.layout import make_layout


def parse_args():
    parser = agp.ArgumentParser()

    parser.add_argument('--result', '-r', help='divik result directory',
                        required=True)
    parser.add_argument('--xy', help='coordinates of points', required=True)

    parser.add_argument('--host', default='127.0.0.1',
                        help='Sets up a host interface to run the visualization'
                             ' on. Use 0.0.0.0 while in Docker.')
    parser.add_argument('--debug', help='enables debug mode', const=True,
                        default=False, action='store_const')

    return parser.parse_args()


def main():
    args = parse_args()
    path = as_divik_result_path(args.result)
    divik_result(path)
    xy(args.xy)
    app.layout = make_layout()
    app.run_server(
        host=args.host,  # required for external access
        debug=args.debug,
        dev_tools_hot_reload=True,
    )


if __name__ == '__main__':
    main()



import json
import os

from divik._inspect.app import divik_result_path


def store_path():
    return os.path.join(divik_result_path(), '.inspect')


def initialize():
    os.makedirs(store_path(), exist_ok=True)


def find_preserved():
    initialize()
    return [
        {
            'label': os.path.splitext(path)[0],
            'value': os.path.splitext(path)[0],
        }
        for path
        in os.listdir(store_path())
    ]


def load(name):
    path = os.path.join(store_path(), '{0}.json'.format(name))
    with open(path) as infile:
        return json.load(infile)


def preserve_by_name(name, disabled, color, level):
    path = os.path.join(store_path(), '{0}.json'.format(name))
    with open(path, 'w') as outfile:
        json.dump({
            'color_overrides': color,
            'disabled_clusters': disabled,
            'level': level
        }, outfile, sort_keys=True)


def restore_level(name):
    return load(name)['level']


def restore_disabled_clusters(name):
    if not name:
        return []
    disabled_clusters = load(name)['disabled_clusters']
    storage = json.loads(disabled_clusters)
    return storage['excluded']


def restore_color_overrides(name):
    return load(name)['color_overrides']



import unittest

from parameterized import parameterized
from sklearn.datasets import make_blobs
from sklearn.metrics import adjusted_rand_score

from divik.cluster._kmeans._core import KMeans
from divik.cluster._kmeans._gap import GAPSearch


def data(n_clusters):
    return make_blobs(n_samples=10000, n_features=2, centers=n_clusters,
                      random_state=0)


class GAPSearchTest(unittest.TestCase):
    @parameterized.expand([
        ("{}_clusters".format(k), k) for k in [1, 2, 3, 4, 7, 8]
    ])
    def test_works_with_gap(self, _, n_clusters):
        X, y = data(n_clusters)
        single_kmeans = KMeans(n_clusters=2)
        kmeans = GAPSearch(single_kmeans, max_clusters=10,
                           sample_size=10000).fit(X)
        rand = adjusted_rand_score(y, kmeans.labels_)
        # allow for misidentification of 1 cluster
        self.assertGreaterEqual(kmeans.n_clusters_ + 1, n_clusters)
        self.assertLessEqual(kmeans.n_clusters_ - 1, n_clusters)
        self.assertGreater(rand, 0.75)

    @parameterized.expand([
        ("{}_clusters".format(k), k) for k in [1, 2, 3, 4, 7, 8]
    ])
    def test_works_with_sampled_gap(self, _, n_clusters):
        X, y = data(n_clusters)
        single_kmeans = KMeans(n_clusters=2)
        kmeans = GAPSearch(single_kmeans, max_clusters=10).fit(X)
        rand = adjusted_rand_score(y, kmeans.labels_)
        # allow for misidentification of 1 cluster
        self.assertGreaterEqual(kmeans.n_clusters_ + 1, n_clusters)
        self.assertLessEqual(kmeans.n_clusters_ - 1, n_clusters)
        self.assertGreater(rand, 0.75)

    def test_works_with_unfit_removal(self):
        n_clusters = 3
        X, y = data(n_clusters)
        single_kmeans = KMeans(n_clusters=2)
        kmeans = GAPSearch(
            single_kmeans, max_clusters=10, drop_unfit=True).fit(X)
        rand = adjusted_rand_score(y, kmeans.labels_)
        # allow for misidentification of 1 cluster
        self.assertGreaterEqual(kmeans.n_clusters_ + 1, n_clusters)
        self.assertLessEqual(kmeans.n_clusters_ - 1, n_clusters)
        self.assertGreater(rand, 0.75)
        self.assertIsNone(kmeans.estimators_)


if __name__ == '__main__':
    unittest.main()



import json

import divik._inspect.color as color

DEFAULT_COLOR = 128


def as_rgb(selected_point, figure):
    try:
        cluster = int(json.loads(selected_point)['cluster'])
    except json.JSONDecodeError:
        return DEFAULT_COLOR, DEFAULT_COLOR, DEFAULT_COLOR
    except ValueError:
        return DEFAULT_COLOR, DEFAULT_COLOR, DEFAULT_COLOR
    colormap = color.make_colormap(figure['data'][0]['z'])
    selected_color = colormap[cluster][1]
    jsonified = selected_color.replace('rgb(', '[').replace(')', ']')
    return json.loads(jsonified)


def update_color_overrides(overrides, new_color, level, selected_point):
    if not overrides:
        overrides = '{}'

    try:
        cluster = str(int(json.loads(selected_point)['cluster']))
    except json.JSONDecodeError:
        return overrides
    except TypeError:
        return overrides

    overrides = json.loads(overrides)
    level = str(level)
    if level not in overrides:
        overrides[level] = {}
    overrides[level][cluster] = new_color
    return json.dumps(overrides)



from typing import (
    Callable,
    Tuple,
    NamedTuple,
    List,
    Optional,
    Union,
)

import numpy as np

Table = np.ndarray  # 2D matrix
Data = Table
Centroids = Table
IntLabels = np.ndarray
SegmentationMethod = Callable[[Data], Tuple[IntLabels, Centroids]]
Clustering = Union['divik.cluster.GAPSearch', 'divik.cluster.DunnSearch']
DivikResult = NamedTuple('DivikResult', [
    ('clustering', Clustering),
    ('feature_selector', 'divik.feature_selection.StatSelectorMixin'),
    ('merged', IntLabels),
    ('subregions', List[Optional['DivikResult']]),
])



import unittest

import numpy as np
import numpy.testing as npt
from sklearn.metrics import accuracy_score

import divik.feature_selection as fs


class GMMSelectorTest(unittest.TestCase):
    def setUp(self):
        np.random.seed(42)
        self.labels = np.concatenate(
            [30 * [0] + 20 * [1] + 30 * [2] + 40 * [3]])
        self.data = self.labels * 5 + np.random.randn(*self.labels.shape)
        self.data = self.data.reshape((1, -1))

    def test_selects_high_component_by_default(self):
        selector = fs.GMMSelector('mean').fit(self.data)
        npt.assert_array_equal(selector.selected_,
                               self.labels == self.labels.max())

    def test_selects_low_component_when_set(self):
        selector = fs.GMMSelector('mean', preserve_high=False).fit(self.data)
        npt.assert_array_equal(selector.selected_,
                               self.labels == self.labels.min())

    def test_that_0_candidates_preserves_all_the_features(self):
        selector = fs.GMMSelector('mean', n_candidates=0).fit(self.data)
        npt.assert_array_equal(selector.selected_, True)

    def test_can_preserve_more_components(self):
        selector = fs.GMMSelector('mean', n_candidates=-1).fit(self.data)
        self.assertGreaterEqual(
            accuracy_score(self.labels >= 2, selector.selected_), 0.99)
        selector = fs.GMMSelector('mean', n_candidates=2).fit(self.data)
        self.assertGreaterEqual(
            accuracy_score(self.labels >= 2, selector.selected_), 0.99)
        selector = fs.GMMSelector('mean', n_candidates=1).fit(self.data)
        self.assertGreaterEqual(
            accuracy_score(self.labels >= 1, selector.selected_), 0.99)
        selector = fs.GMMSelector('mean', n_candidates=1, preserve_high=False)
        selector.fit(self.data)
        self.assertGreaterEqual(
            accuracy_score(self.labels < 3, selector.selected_), 0.99)

    def test_fails_log_with_negative_features(self):
        self.data[0, 1] = -1
        selector = fs.GMMSelector('mean', use_log=True)
        with self.assertRaises(ValueError):
            selector.fit(self.data)

    def test_works_for_mean_and_var_only(self):
        fs.GMMSelector('mean')
        fs.GMMSelector('var')
        with self.assertRaises(ValueError):
            fs.GMMSelector('yolo')

    def test_preserves_min_features_precisely_or_rate(self):
        selector = fs.GMMSelector('mean', min_features=50).fit(self.data)
        self.assertGreaterEqual(selector.selected_.sum(), 50)
        selector = fs.GMMSelector('mean', min_features_rate=0.5).fit(self.data)
        self.assertGreaterEqual(selector.selected_.sum(),
                                0.5 * self.labels.size)



#!/usr/bin/env python
from functools import partial
import logging
import os
import pickle
from typing import List, Tuple

import numpy as np
import skimage.io as sio

from divik.cluster import GAPSearch, KMeans
import divik._cli._utils as scr
import divik.core as u


Segmentations = List[Tuple[u.IntLabels, u.Centroids]]


def get_segmentations(kmeans: GAPSearch) -> Segmentations:
    return [(est.labels_, est.cluster_centers_) for est in kmeans.estimators_]


def make_segmentations_matrix(kmeans: GAPSearch) -> np.ndarray:
    return np.hstack([e.labels_.reshape(-1, 1) for e in kmeans.estimators_])


def save(kmeans: GAPSearch, destination: str, xy: np.ndarray=None):
    logging.info("Saving result.")

    logging.info("Saving model.")
    fname = partial(os.path.join, destination)
    with open(fname('model.pkl'), 'wb') as pkl:
        pickle.dump(kmeans, pkl)

    logging.info("Saving segmentations.")

    np.savetxt(fname('final_partition.csv'), kmeans.labels_.reshape(-1, 1),
               delimiter=', ', fmt='%i')
    np.save(fname('final_partition.npy'), kmeans.labels_.reshape(-1, 1))

    segmentations = get_segmentations(kmeans)
    with open(fname('segmentations.pkl'), 'wb') as pkl:
        pickle.dump(segmentations, pkl)

    partitions = make_segmentations_matrix(kmeans)
    np.savetxt(fname('partitions.csv'), partitions, delimiter=', ', fmt='%i')

    for i in range(partitions.shape[1]):
        np.savetxt(fname('partitions.{0}.csv').format(i + kmeans.min_clusters),
                   partitions[:, i].reshape(-1, 1), delimiter=', ', fmt='%i')

    if xy is not None:
        visualization = u.visualize(kmeans.labels_, xy=xy)
        sio.imsave(fname('partitions.{0}.png').format(kmeans.n_clusters_),
                   visualization)


def main():
    data, config, destination, xy = scr.initialize()
    try:
        single_kmeans = KMeans(**config['kmeans'])
        kmeans = GAPSearch(single_kmeans, **config['gap'])
        kmeans.fit(data)
    except Exception as ex:
        logging.error("Failed with exception.")
        logging.error(repr(ex))
        raise
    save(kmeans, destination, xy)


if __name__ == '__main__':
    main()



from dash.dependencies import Output, Input, State

from divik._inspect.app import app
from divik._inspect.layout import Fields


def make_tab_callback(tab_id):
    @app.callback(
        Output(tab_id, 'style'),
        [Input(Fields.MENU_TABS, 'value')],
        [State(tab_id, 'style')]
    )
    def update_visibility(value, style):
        style = style or {}
        if value == tab_id:
            style['display'] = 'block'
        else:
            style['display'] = 'none'
        return style


make_tab_callback(Fields.COLOR_MANIPULATION_TAB)
make_tab_callback(Fields.EXCLUSION_TAB)
make_tab_callback(Fields.PERSISTENCE_TAB)



import colorsys
import itertools
from fractions import Fraction

import numpy as np


def _get_fractions():
    """
    [Fraction(0, 1), Fraction(1, 2), Fraction(1, 4), Fraction(3, 4), Fraction(1, 8), Fraction(3, 8), Fraction(5, 8), Fraction(7, 8), Fraction(1, 16), Fraction(3, 16), ...]
    [0.0, 0.5, 0.25, 0.75, 0.125, 0.375, 0.625, 0.875, 0.0625, 0.1875, ...]
    """
    yield 0
    for k in itertools.count():
        i = 2 ** k
        for j in range(1, i, 2):
            yield Fraction(j, i)


_DEFAULT_SATURATIONS = [Fraction(6, 10)]
_DEFAULT_VALUES = [Fraction(8, 10), Fraction(5, 10)]


def _generate_hsv_neighbourhood(h):
    for s in _DEFAULT_SATURATIONS:
        for v in _DEFAULT_VALUES:
            yield (h, s, v)


def as_colormap_color(h, s, v):
    r, g, b = colorsys.hsv_to_rgb(h, s, v)
    r, g, b = int(255 * r), int(255 * g), int(255 * b)
    return 'rgb({0},{1},{2})'.format(r, g, b)


def colormap():
    fractions = _get_fractions()
    neighbors = map(_generate_hsv_neighbourhood, fractions)
    neighbors = itertools.chain.from_iterable(neighbors)
    for args in neighbors:
        yield as_colormap_color(*args)


_DISABLED_COLOR = 'rgb(128, 128, 128)'


def _pick_color(value, color, disabled, overrides):
    if value in disabled:
        return _DISABLED_COLOR
    if str(value) in overrides:
        return overrides[str(value)]
    return color


def make_colormap(values, disabled=None, overrides=None):
    if disabled is None:
        disabled = []
    if overrides is None:
        overrides = {}
    values = np.array(values)
    return [
        [
            value / values.max(),
            _pick_color(value, color, disabled, overrides)
        ]
        for value, color
        in zip(np.unique(values), colormap())
    ]



import unittest

import numpy as np
from sklearn.datasets import make_blobs

import divik.cluster as km
import divik.score as sc


def dummy_kmeans(labels, cluster_centers):
    kmeans = km.KMeans(n_clusters=len(cluster_centers), distance='euclidean',
                       init='extreme', max_iter=10)
    kmeans.labels_ = labels
    kmeans.cluster_centers_ = cluster_centers
    return kmeans


class TestGap(unittest.TestCase):
    def setUp(self):
        self.X, _ = make_blobs(n_samples=10000, n_features=2, centers=3,
                               random_state=0)
        self.kmeans_3 = km.KMeans(n_clusters=3).fit(self.X)
        self.kmeans_7 = km.KMeans(n_clusters=7).fit(self.X)

    def test_computes_score(self):
        score = sc.gap(self.X, self.kmeans_3)
        self.assertFalse(np.isnan(score))

    def test_good_labeling_has_greater_score(self):
        better = sc.gap(self.X, self.kmeans_3)
        worse = sc.gap(self.X, self.kmeans_7)
        self.assertGreater(better, worse)

    def test_returns_std_if_requested(self):
        gap, std = sc.gap(self.X, self.kmeans_3, return_deviation=True)
        self.assertIsNotNone(gap)
        self.assertIsNotNone(std)



import json

from dash.dependencies import Output, Input, State

from divik._inspect import recolor as recolor, persistence as per
from divik._inspect.app import app
from divik._inspect.layout import Fields


@app.callback(
    Output(Fields.SELECTED_POINT, 'children'),
    [Input(Fields.CLUSTERS_GRAPH, 'clickData')]
)
def update_click_data(click_data):
    if click_data:
        point = click_data['points'][0]
        return json.dumps({
            'cluster': point['z'] if 'z' in point else 'outside the sample',
            'x': point['x'],
            'y': point['y']
        }, indent=4, sort_keys=True)
    return 'no point selected'


@app.callback(
    Output(Fields.CLUSTER_COLOR_SAMPLE, 'style'),
    [
        Input(Fields.CLUSTER_COLOR_R, 'value'),
        Input(Fields.CLUSTER_COLOR_G, 'value'),
        Input(Fields.CLUSTER_COLOR_B, 'value'),
    ]
)
def update_color_sample(r, g, b):
    return {
        'background-color': 'rgb({0}, {1}, {2})'.format(r, g, b),
        'height': '1em'
    }


@app.callback(
    Output(Fields.CLUSTER_COLOR_R, 'value'),
    [Input(Fields.SELECTED_POINT, 'children')],
    [State(Fields.CLUSTERS_GRAPH, 'figure')]
)
def set_r_from_point(selected_point, figure):
    return recolor.as_rgb(selected_point, figure)[0]


@app.callback(
    Output(Fields.CLUSTER_COLOR_G, 'value'),
    [Input(Fields.SELECTED_POINT, 'children')],
    [State(Fields.CLUSTERS_GRAPH, 'figure')]
)
def set_g_from_point(selected_point, figure):
    return recolor.as_rgb(selected_point, figure)[1]


@app.callback(
    Output(Fields.CLUSTER_COLOR_B, 'value'),
    [Input(Fields.SELECTED_POINT, 'children')],
    [State(Fields.CLUSTERS_GRAPH, 'figure')]
)
def set_b_from_point(selected_point, figure):
    return recolor.as_rgb(selected_point, figure)[2]


@app.callback(
    Output(Fields.COLOR_OVERRIDES_STORAGE, 'children'),
    [
        Input(Fields.CLUSTER_COLOR_APPLY, 'n_clicks_timestamp'),
        Input(Fields.CLUSTER_COLOR_RESET, 'n_clicks_timestamp'),
        Input(Fields.LOAD_PROFILE, 'n_clicks_timestamp'),
    ],
    [
        State(Fields.COLOR_OVERRIDES_STORAGE, 'children'),
        State(Fields.CLUSTER_COLOR_SAMPLE, 'style'),
        State(Fields.LEVEL, 'value'),
        State(Fields.SELECTED_POINT, 'children'),
        State(Fields.SAVED_PROFILES, 'value')
    ]
)
def store_color_override(apply, reset, load, overrides, sample_style, level,
                         selected_point, name):
    apply = apply or 0
    reset = reset or 0
    load = load or 0
    last_click = max(apply, load, reset)
    if not last_click or reset == last_click:
        return '{}'
    if load == last_click:
        if name:
            return per.restore_color_overrides(name)
        return '{}'
    if apply == last_click:
        return recolor.update_color_overrides(
            overrides, sample_style['background-color'], level, selected_point)
    raise NotImplementedError('Unhandled case.')



"""Implementation of PCA with data-driven variance explanation limit"""
import warnings

from kneed import KneeLocator
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.decomposition import PCA

from divik.core import configurable


def knee(explained_variance) -> int:
    """Find empirical knee point for explained variance"""
    xaxis = np.arange(explained_variance.size, dtype=int)
    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            limit = KneeLocator(x=xaxis,
                                y=explained_variance,
                                S=1.,
                                direction='increasing',
                                curve='concave').knee
    except IndexError:  # This is needed for kneed >= 0.5.3
        limit = None
    if limit is not None:
        return limit
    return explained_variance.size


@configurable
class KneePCA(BaseEstimator, TransformerMixin):
    """Principal component analysis (PCA) with knee method

    PCA with automated components selection based on knee method
    over cumulative explained variance. Remaining components are
    discarded.

    Parameters
    -----------
    whiten : bool, optional (default False)
        When True (False by default) the `pca_.components_` vectors are
        multiplied by the square root of n_samples and then divided by the
        singular values to ensure uncorrelated outputs with unit
        component-wise variances.

        Whitening will remove some information from the transformed signal
        (the relative variance scales of the components) but can sometime
        improve the predictive accuracy of the downstream estimators by
        making their data respect some hard-wired assumptions.

    refit : bool, optional (default False)
        When True (False by default) the `pca_` is re-fit with the smaller
        number of components. This could reduce memory footprint, but
        requires training fitting PCA.

    Attributes
    ----------
    pca_ : PCA
        Fit PCA estimator.

    n_components_ : int
        The number of selected components.
    """
    def __init__(self, whiten: bool = False, refit: bool = False):
        self.whiten = whiten
        self.refit = refit

    def fit(self, X, y=None):
        """Fit the model from data in X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training vector, where n_samples is the number of samples
            and n_features is the number of features.

        Y: Ignored.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        # Note: random_state is not used in this config!
        self.pca_ = PCA(n_components=None, copy=True, whiten=self.whiten,
                        svd_solver='full', tol=0.0, iterated_power='auto',
                        random_state=None).fit(X)
        self.n_components_ = knee(
            np.cumsum(self.pca_.explained_variance_ratio_))
        if self.refit:
            self.pca_ = PCA(n_components=self.n_components_, copy=True,
                            whiten=self.whiten, svd_solver='full', tol=0.0,
                            iterated_power='auto', random_state=None).fit(X)
        return self

    def transform(self, X, y=None):
        """Apply dimensionality reduction to X.

        X is projected on the first principal components previously extracted
        from a training set.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            New data, where n_samples is the number of samples
            and n_features is the number of features.

        Returns
        -------
        X_new : array-like, shape (n_samples, n_components)

        Examples
        --------

        >>> import numpy as np
        >>> from divik.feature_extraction import KneePCA
        >>> X = np.array([[-1, -1], [-2, -1], [-3, -2], [1, 1], [2, 1], [3, 2]])
        >>> pca = KneePCA(refit=True)
        >>> pca.fit(X)
        KneePCA(refit=True)
        >>> pca.transform(X) # doctest: +SKIP
        """
        loads = self.pca_.transform(X)
        return loads[:, :self.n_components_]

    def inverse_transform(self, X):
        """Transform data back to its original space.

        In other words, return an input X_original whose transform would be X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_components)
            New data, where n_samples is the number of samples
            and n_components is the number of components.

        Returns
        -------
        X_original array-like, shape (n_samples, n_features)

        Notes
        -----
        If whitening is enabled, inverse_transform will compute the
        exact inverse operation, which includes reversing whitening.
        """
        n_missing = self.pca_.n_components_ - self.n_components_
        missing = np.zeros((X.shape[0], n_missing))
        full = np.hstack([X, missing])
        return self.pca_.inverse_transform(full)



import unittest

import numpy as np

import divik.feature_selection as fs


class HighAbundanceAndVarianceSelectorTest(unittest.TestCase):
    def setUp(self):
        np.random.seed(42)
        self.labels = np.concatenate(
            [30 * [0] + 20 * [1] + 30 * [2] + 40 * [3]])
        self.data = np.vstack(100 * [self.labels * 10.])
        self.data += np.random.randn(*self.data.shape)
        sub = self.data[:, :-40]
        sub += 5 * np.random.randn(*sub.shape)

    def test_discards_low_abundance(self):
        selector = fs.HighAbundanceAndVarianceSelector().fit(self.data)
        TNR = (selector.selected_[self.labels == 0] == False).mean()
        self.assertGreaterEqual(TNR, 0.99)

    def test_discards_low_variance(self):
        selector = fs.HighAbundanceAndVarianceSelector().fit(self.data)
        TNR = (selector.selected_[self.labels == 3] == False).mean()
        self.assertGreaterEqual(TNR, 0.99)

    def test_passes_informative(self):
        selector = fs.HighAbundanceAndVarianceSelector().fit(self.data)
        expected = np.logical_or(self.labels == 1, self.labels == 2)
        TPR = selector.selected_[expected].mean()
        self.assertGreaterEqual(TPR, 0.99)


class OutlierAbundanceAndVarianceSelectorTest(unittest.TestCase):
    def setUp(self):
        np.random.seed(42)
        self.labels = np.concatenate(
            [6 * [0] + 5 * [1] + 25 * [2] + 50 * [3] + 3 * [4] + 4 * [5]])
        self.data = np.vstack(1000 * [self.labels * 10.])
        self.data += np.random.randn(*self.data.shape)
        self.outlier_mean = np.logical_or(self.labels == 4, self.labels == 5)
        self.high_var = np.zeros_like(self.outlier_mean)
        self.high_var[np.where(self.labels == 2)[0][:5]] = True
        shape = self.data[:, self.high_var].shape
        self.data[:, self.high_var] += 5 * np.random.randn(*shape)

    def test_discards_outlier_abundance(self):
        selector = fs.OutlierAbundanceAndVarianceSelector().fit(self.data)
        TNR = (selector.selected_[self.outlier_mean] == False).mean()
        self.assertGreaterEqual(TNR, 0.95)

    def test_discards_outlier_variance(self):
        selector = fs.OutlierAbundanceAndVarianceSelector().fit(self.data)
        TNR = (selector.selected_[self.high_var] == False).mean()
        self.assertGreaterEqual(TNR, 0.95)

    def test_selects_percentage_of_features(self):
        N = self.labels.size
        p = 5. / N
        selector = fs.OutlierAbundanceAndVarianceSelector(p=p).fit(self.data)
        self.assertAlmostEqual(selector.selected_.mean(), p, places=2)
        p = 20. / N
        selector = fs.OutlierAbundanceAndVarianceSelector(p=p).fit(self.data)
        self.assertAlmostEqual(selector.selected_.mean(), p, places=2)



"""Access to MATLAB legacy functionalities."""
import gamred_native as gn
import numpy as np


def find_thresholds(values: np.ndarray, max_components: int = 10) \
        -> np.ndarray:
    """Find candidate thresholds for decomposition of values by GMM.

    Parameters
    ----------
    values:
        vector of values to decompose

    max_components:
        maximal number of components to decompose into

    Returns
    -------
    array of candidate thresholds from crossings between GMM components
    """
    if values.size <= max_components:
        max_components = values.size
    if values.size == 0:
        return np.array([])
    if max_components <= 0:
        raise ValueError("max_components must be positive")
    if values.ndim != 1:
        raise ValueError("values must be 1D array")
    values = np.ascontiguousarray(values)
    offset = np.min(values)
    if np.min(values) == np.max(values):
        return np.array([])
    return gn.find_thresholds(values - offset, max_components) + offset



import unittest

from parameterized import parameterized
from sklearn.datasets import make_blobs
from sklearn.metrics import adjusted_rand_score

from divik.cluster._kmeans._core import KMeans
from divik.cluster._kmeans._dunn import DunnSearch


# if the std is too big, it won't work with high number of clusters,
# because the 20x20 grid is quite small for 10 000 points
def data(n_clusters):
    return make_blobs(n_samples=10000, n_features=2, centers=n_clusters,
                      random_state=0, cluster_std=0.1)


class DunnSearchTest(unittest.TestCase):
    # Simplified Dunn favorizes small number of clusters.
    @parameterized.expand([
        ("{}_clusters".format(k), k) for k in [2, 3, 4]
    ])
    def test_works_with_simplified_dunn(self, _, n_clusters):
        X, y = data(n_clusters)
        single_kmeans = KMeans(n_clusters=2, init='kdtree')
        kmeans = DunnSearch(single_kmeans, max_clusters=10).fit(X)
        rand = adjusted_rand_score(y, kmeans.labels_)
        self.assertEqual(n_clusters, kmeans.n_clusters_)
        self.assertGreater(rand, 0.75)

    @parameterized.expand([
        ("{}_clusters".format(k), k) for k in range(2, 9)
    ])
    def test_works_with_full_exact_dunn(self, _, n_clusters):
        X, y = data(n_clusters)
        single_kmeans = KMeans(n_clusters=2, init='kdtree')
        kmeans = DunnSearch(single_kmeans, max_clusters=15,
                            inter='closest', intra='furthest',
                            ).fit(X)
        rand = adjusted_rand_score(y, kmeans.labels_)
        self.assertEqual(n_clusters, kmeans.n_clusters_)
        self.assertGreater(rand, 0.75)

    @parameterized.expand([
        ("{}_clusters".format(k), k) for k in range(2, 9)
    ])
    def test_works_with_sampled_exact_dunn(self, _, n_clusters):
        X, y = data(n_clusters)
        single_kmeans = KMeans(n_clusters=2, init='kdtree')
        kmeans = DunnSearch(single_kmeans, max_clusters=15, method='sampled',
                            inter='closest', intra='furthest', n_trials=10,
                            ).fit(X)
        rand = adjusted_rand_score(y, kmeans.labels_)
        self.assertEqual(n_clusters, kmeans.n_clusters_)
        self.assertGreater(rand, 0.75)
    
    def test_works_with_unfit_removal(self):
        n_clusters = 3
        X, y = data(n_clusters)
        single_kmeans = KMeans(n_clusters=2)
        kmeans = DunnSearch(
            single_kmeans, max_clusters=10, drop_unfit=True).fit(X)
        rand = adjusted_rand_score(y, kmeans.labels_)
        self.assertEqual(n_clusters, kmeans.n_clusters_)
        self.assertGreater(rand, 0.75)
        self.assertIsNone(kmeans.estimators_)


if __name__ == '__main__':
    unittest.main()



import argparse as agp

from skimage.io import imsave

from divik._cli._data_io import load_data
from divik.core import visualize


def parse_args():
    parser = agp.ArgumentParser()

    parser.add_argument('--labels', required=True,
                        help='Path to file with labels')
    parser.add_argument('--xy', required=True,
                        help='Path to file with spatial coordinates')
    parser.add_argument('--destination', required=True,
                        help='Path to destination file')

    return parser.parse_args()


def main():
    args = parse_args()
    print("Loading {0} as labels.".format(args.labels))
    print("Loading {0} as coordinates.".format(args.xy))
    print("Labels will be saved to {0}.".format(args.destination))
    xy = load_data(args.xy).astype(int)
    labels = load_data(args.labels).astype(int)
    visualization = visualize(label=labels, xy=xy)
    imsave(args.destination, visualization)


if __name__ == '__main__':
    main()



from enum import Enum, unique

import dash_core_components as dcc
import dash_html_components as html

from divik._summary import depth
from divik._inspect.app import divik_result, divik_result_path
from divik._inspect.figure import default_clusters_figure


def _as_id(name: str) -> str:
    return name.lower().replace('_', '-')


_FIELDS = [
    'TITLE',
    'CLUSTERS_CONTAINER',
    'CLUSTERS_GRAPH',
    'LEVEL',
    'DISABLED_CLUSTERS_PICKER',
    'DISABLED_CLUSTERS_STORAGE',
    'SELECTED_POINT',
    'COLOR_OVERRIDES_STORAGE',
    'CLUSTER_COLOR_R',
    'CLUSTER_COLOR_G',
    'CLUSTER_COLOR_B',
    'CLUSTER_COLOR_SAMPLE',
    'CLUSTER_COLOR_APPLY',
    'CLUSTER_COLOR_RESET',
    'SAVED_PROFILES',
    'LOAD_PROFILE',
    'NEW_PROFILE_NAME',
    'SAVE_PROFILE',
    'MENU_TABS',
    'COLOR_MANIPULATION_TAB',
    'EXCLUSION_TAB',
    'PERSISTENCE_TAB',
]


# noinspection PyArgumentList
Fields = unique(Enum(
    'Fields', [(name, _as_id(name)) for name in _FIELDS], type=str))


def make_title():
    return html.H4(id=Fields.TITLE,
                   children=divik_result_path(),
                   style={'text-align': 'center'})


def make_plot():
    result_depth = depth(divik_result())
    return html.Div(id=Fields.CLUSTERS_CONTAINER, children=[
        dcc.Graph(id=Fields.CLUSTERS_GRAPH,
                  figure=default_clusters_figure(),
                  style={'min-height': 600}),
        html.H4('Level'),
        dcc.Slider(id=Fields.LEVEL, value=1, min=1, max=result_depth - 1,
                   step=1, marks={i: i for i in range(1, result_depth)})
    ], className='eight columns')


def make_selected_point_display():
    return html.Div([
        html.H4('Selected point'),
        html.Code(id=Fields.SELECTED_POINT)
    ])


def make_new_color_assignment_panel():
    return html.Div([
        html.H4('Assign new cluster color'),
        html.Div(id=Fields.CLUSTER_COLOR_SAMPLE),
        dcc.Slider(id=Fields.CLUSTER_COLOR_R, min=0, max=255, value=128),
        dcc.Slider(id=Fields.CLUSTER_COLOR_G, min=0, max=255, value=128),
        dcc.Slider(id=Fields.CLUSTER_COLOR_B, min=0, max=255, value=128),
        html.Button('Apply', id=Fields.CLUSTER_COLOR_APPLY),
    ])


def make_colors_reset_panel():
    return html.Div([
        html.H4('Reset clusters colors'),
        html.Button('Reset', id=Fields.CLUSTER_COLOR_RESET),
    ])


def make_disabled_clusters_panel():
    return html.Div([
        html.H4('Disabled clusters'),
        dcc.Dropdown(id=Fields.DISABLED_CLUSTERS_PICKER, multi=True)
    ])


def make_load_profile_panel():
    return html.Div([
        html.H4('Load profile'),
        dcc.Dropdown(id=Fields.SAVED_PROFILES),
        html.Br(),
        html.Button('Load', id=Fields.LOAD_PROFILE)
    ])


def make_save_profile_panel():
    return html.Div([
        html.H4('Save profile'),
        dcc.Input(id=Fields.NEW_PROFILE_NAME, type='text',
                  style={'display': 'block'}),
        html.Br(),
        html.Button('Save', id=Fields.SAVE_PROFILE)
    ])


def make_color_manipulation_tab():
    return html.Div(id=Fields.COLOR_MANIPULATION_TAB, children=[
        make_selected_point_display(),
        make_new_color_assignment_panel(),
        make_colors_reset_panel(),
    ])


def make_exclusion_tab():
    return html.Div(id=Fields.EXCLUSION_TAB, children=[
        make_disabled_clusters_panel(),
    ])


def make_persistence_tab():
    return html.Div(id=Fields.PERSISTENCE_TAB, children=[
        make_load_profile_panel(),
        make_save_profile_panel(),
    ])


def make_menu_panel():
    return html.Div(children=[
        dcc.Tabs(id=Fields.MENU_TABS,
                 value=Fields.COLOR_MANIPULATION_TAB,
                 children=[
                     dcc.Tab(label='COLOR',
                             value=Fields.COLOR_MANIPULATION_TAB),
                     dcc.Tab(label='SELECT',
                             value=Fields.EXCLUSION_TAB),
                     dcc.Tab(label='PROFILE',
                             value=Fields.PERSISTENCE_TAB),
                 ]),
        make_color_manipulation_tab(),
        make_exclusion_tab(),
        make_persistence_tab(),
    ], className='three columns')


def make_storage():
    return html.Div([
        html.Div(id=Fields.DISABLED_CLUSTERS_STORAGE, style={'display': 'none'}),
        html.Div(id=Fields.COLOR_OVERRIDES_STORAGE, style={'display': 'none'})
    ])


def make_layout():
    return html.Div([
        make_title(),
        make_plot(),
        make_menu_panel(),
        make_storage(),
    ])



from abc import ABCMeta
import logging

import numpy as np
from sklearn.base import BaseEstimator
try:
    from sklearn.feature_selection._base import SelectorMixin
except ModuleNotFoundError:
    from sklearn.feature_selection.base import SelectorMixin

from divik.core import configurable


class StatSelectorMixin(SelectorMixin, metaclass=ABCMeta):
    """
    Transformer mixin that performs feature selection given a support mask

    This mixin provides a feature selector implementation with `transform` and
    `inverse_transform` functionality given that `selected_` is specified
    during `fit`.

    Additionally, provides a `_to_characteristics` and `_to_raw` implementations
    given `stat`, optionally `use_log` and `preserve_high`.
    """
    def _to_characteristics(self, X):
        """Extract & normalize characteristics from data"""
        if self.stat == 'mean':
            vals = np.mean(X, axis=0)
        elif self.stat == 'var':
            vals = np.var(X, axis=0)
        else:
            logging.error('stat must be one of {"mean", "var"}')
            raise ValueError('stat must be one of {"mean", "var"}')

        if hasattr(self, 'use_log') and self.use_log:
            if np.any(vals < 0):
                logging.error("Feature characteristic cannot be negative "
                              "with log filtering")
                raise ValueError("Feature characteristic cannot be negative "
                                 "with log filtering")
            vals = np.log(vals)

        if hasattr(self, 'preserve_high') and not self.preserve_high:
            vals = -vals

        return vals

    def _to_raw(self, threshold):
        """Convert threshold to the feature characteristic space"""
        if hasattr(self, 'preserve_high') and not self.preserve_high:
            threshold = -threshold
        if hasattr(self, 'use_log') and self.use_log:
            threshold = np.exp(threshold)
        return threshold

    def _get_support_mask(self):
        """
        Get the boolean mask indicating which features are selected

        Returns
        -------
        support : boolean array of shape [# input features]
            An element is True iff its corresponding feature is selected for
            retention.
        """
        return self.selected_


@configurable
class NoSelector(BaseEstimator, StatSelectorMixin):
    """Dummy selector to use when no selection is supposed to be made."""
    def __init__(self):
        pass

    def fit(self, X, y=None):
        """Pass data forward

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            Sample vectors to pass.

        y : any
            Ignored. This parameter exists only for compatibility with
            sklearn.pipeline.Pipeline.

        Returns
        -------
        self
        """
        self.selected_ = np.ones((X.shape[1],), dtype=bool)
        return self



from typing import NamedTuple

from kneed import KneeLocator
import numpy as np


FeaturesSelection = NamedTuple('FeaturesSelection', [
    ('score', np.ndarray),
    ('index', np.ndarray),
    ('selection', np.ndarray),
    ('threshold', float)
])


def _gradient(values: np.ndarray, order: int=1) -> np.ndarray:
    result = values
    for _ in range(order):
        result = np.gradient(result)
    return result


def _plateau_point(sorted_scores) -> int:
    gradient = np.abs(_gradient(sorted_scores))
    is_plateau = gradient <= np.percentile(gradient, 1)
    plateau_points = np.nonzero(is_plateau)[0]
    return int(np.median(plateau_points))


def _knee_point(decreasing_segment: np.ndarray) -> int:
    locator = KneeLocator(x=np.arange(decreasing_segment.size, dtype=int),
                          y=decreasing_segment,
                          S=1.,
                          curve='convex',
                          direction='decreasing')
    assert locator.knee is not None
    return locator.knee


def select_features(feature_scores) -> FeaturesSelection:
    feature_scores = feature_scores.ravel()
    index = np.argsort(-feature_scores)
    score = feature_scores[index]
    plateau_point = _plateau_point(score)
    concave_up_segment = score[:plateau_point]
    knee_location = _knee_point(concave_up_segment)
    threshold = score[knee_location]
    selection = feature_scores >= threshold
    return FeaturesSelection(score, index, selection, threshold)



import unittest

from scipy.spatial.distance import cdist
from sklearn.datasets import make_blobs

from divik.sampler._uniform_sampler import UniformSampler, UniformPCASampler


def data():
    return make_blobs(n_samples=1000, n_features=30, centers=5, random_state=1)


class UniformSamplerTest(unittest.TestCase):
    def test_keeps_shapes(self):
        X, _ = data()
        sampler = UniformSampler().fit(X)
        self.assertEqual(sampler.get_sample(0).shape, X.shape)
        sampler = UniformSampler(n_rows=10).fit(X)
        self.assertEqual(sampler.get_sample(0).shape, (10, X.shape[1]))

    def test_keeps_limits(self):
        X, _ = data()
        sampler = UniformSampler().fit(X)
        X_rand = sampler.get_sample(0)
        self.assertTrue((X.min() <= X_rand.min()).all())
        self.assertTrue((X.max() >= X_rand.max()).all())


class UniformPCASamplerTest(unittest.TestCase):
    def test_keeps_shapes(self):
        X, _ = data()
        sampler = UniformPCASampler().fit(X)
        self.assertEqual(sampler.get_sample(0).shape, X.shape)
        sampler = UniformPCASampler(n_rows=10).fit(X)
        self.assertEqual(sampler.get_sample(0).shape, (10, X.shape[1]))

    def test_keeps_correlation(self):
        X, _ = data()
        sampler = UniformPCASampler().fit(X)
        X_rand = sampler.get_sample(0)
        d = float(cdist([X.mean(axis=0)], [X_rand.mean(axis=0)],
                        metric='correlation'))
        self.assertLess(d, 0.15)


if __name__ == '__main__':
    unittest.main()



"""gin-config compatibility"""
import os
import sys

try:
    import gin
    from absl import flags
    import divik.core._gin_bugfix
    _HAS_GIN = True
except ImportError:
    _HAS_GIN = False


MISSING_GIN_ERROR = """
gin-config package missing. You should install divik with appropriate extras:
    pip install divik[gin]
"""


def parse_gin_args():
    """Parse arguments with gin-config

    If you have `gin` extras installed, you can call `parse_gin_args`
    to parse command line arguments or config files to configure
    your runs.

    Command line arguments are used like `--param='DiviK.k_max=50'`.
    Config files are passed via `--config=path.gin`.

    More about format of `.gin` files can be found here:
    https://github.com/google/gin-config
    """
    try:
        import gin
        from absl import flags
    except ImportError as ex:
        raise ImportError(MISSING_GIN_ERROR) from ex
    flags.DEFINE_multi_string(
        'config', None, 'List of paths to the config files.')
    flags.DEFINE_multi_string(
        'param', None, 'Newline separated list of Gin parameter bindings.')
    FLAGS = flags.FLAGS
    FLAGS(sys.argv)
    gin.parse_config_files_and_bindings(FLAGS.config, FLAGS.param)


def dump_gin_args(destination):
    """Dump gin-config effective configuration

    If you have `gin` extras installed, you can call `dump_gin_args`
    save effective gin configuration to a file.
    """
    try:
        import gin
    except ImportError as ex:
        raise ImportError(MISSING_GIN_ERROR) from ex
    with open(os.path.join(destination, 'config.gin'), 'w') as outfile:
        outfile.write(gin.operative_config_str())


if _HAS_GIN:
    configurable = gin.configurable
else:
    def configurable(name_or_fn=None, *args, **kwargs):
        if name_or_fn is None:
            return lambda x: x
        return name_or_fn



import unittest
from unittest.mock import patch
import os

import divik.core as u


@patch.object(os, os.cpu_count.__name__, new=lambda: 4)
class NJobsTest(unittest.TestCase):
    def test_gets_n_for_n_lower_than_cpus(self):
        self.assertEqual(u.get_n_jobs(3), 3)

    def test_gets_n_for_n_equal_cpus(self):
        self.assertEqual(u.get_n_jobs(4), 4)

    def test_gets_all_for_minus_one(self):
        self.assertEqual(u.get_n_jobs(-1), 4)

    def test_gets_all_but_n_for_negative(self):
        self.assertEqual(u.get_n_jobs(-2), 3)

    def test_gets_all_for_zero(self):
        self.assertEqual(u.get_n_jobs(0), 4)


class Dummy:
    def __init__(self, a, b = 5):
        self.a = a
        self.b = b


class BuildTest(unittest.TestCase):
    def test_builds_instance_with_proper_args(self):
        dummy = u.build(Dummy, a=3, b=6)
        self.assertEqual(dummy.a, 3)
        self.assertEqual(dummy.b, 6)

        dummy = u.build(Dummy, a=3)
        self.assertEqual(dummy.a, 3)
        self.assertEqual(dummy.b, 5)
    
    def test_builds_with_too_many_args(self):
        dummy = u.build(Dummy, a=3, c=4)
        self.assertEqual(dummy.a, 3)
        self.assertEqual(dummy.b, 5)



import unittest

import numpy as np
from sklearn.datasets import make_blobs

import divik.cluster as km
from divik.score._dunn import (
    _inter_centroid,
    _inter_closest,
    _intra_avg,
    _intra_furthest,
    dunn,
    sampled_dunn,
)


class DummyKMeans:
    n_clusters = 2
    cluster_centers_ = np.array([[2], [5]])
    labels_ = np.array([0, 0, 1, 1], dtype=int)
    distance = 'euclidean'


class TestDunnDistances(unittest.TestCase):
    def test_inter_centroid(self):
        data = np.array([[1], [3], [4], [6]])
        dst = _inter_centroid(DummyKMeans(), data)
        self.assertAlmostEqual(dst, 3.)

    def test_inter_closest(self):
        data = np.array([[1], [3], [4], [6]])
        dst = _inter_closest(DummyKMeans(), data)
        self.assertAlmostEqual(dst, 1.)

    def test_intra_avg(self):
        data = np.array([[1], [3], [4], [6]])
        dst = _intra_avg(DummyKMeans(), data)
        self.assertAlmostEqual(dst, 1.)

    def test_intra_furthest(self):
        data = np.array([[1], [3], [4], [6]])
        dst = _intra_furthest(DummyKMeans(), data)
        self.assertAlmostEqual(dst, 2.)


class TestDunn(unittest.TestCase):
    def test_computes_inter_to_intracluster_distances_rate(self):
        data = np.array([[1], [3], [4], [6]])
        dunn_ = dunn(DummyKMeans(), data)
        self.assertAlmostEqual(dunn_, 3.)

    def test_works_with_other_distances(self):
        data = np.array([[1], [3], [4], [6]])
        dunn_ = dunn(DummyKMeans(), data, inter='closest', intra='furthest')
        self.assertAlmostEqual(dunn_, .5)


class TestSamplingDunn(unittest.TestCase):
    def setUp(self):
        self.X, _ = make_blobs(n_samples=10000, n_features=2, centers=3,
                               random_state=0, cluster_std=0.1)
        self.kmeans_3 = km.KMeans(n_clusters=3).fit(self.X)
        self.kmeans_7 = km.KMeans(n_clusters=7).fit(self.X)

    def test_computes_score(self):
        score = sampled_dunn(self.kmeans_3, self.X)
        self.assertFalse(np.isnan(score))

    def test_good_labeling_has_top_score(self):
        kmeans = [km.KMeans(n_clusters=k, init='kdtree').fit(self.X)
                  for k in range(2, 11)]
        dunn_ = [sampled_dunn(mdl, self.X) for mdl in kmeans]
        best = int(np.argmax(dunn_))
        self.assertEqual(kmeans[best].n_clusters, self.kmeans_3.n_clusters)



from sklearn.base import BaseEstimator

from divik.core import configurable
from ._exims import exims
from ._selection import select_features

from divik.feature_selection._stat_selector_mixin import SelectorMixin


@configurable
class EximsSelector(BaseEstimator, SelectorMixin):
    """Select features based on their spatial distribution

    Preserves features that yield biologically plausible structures.

    References
    ----------

    Wijetunge, Chalini D., et al. "EXIMS: an improved data analysis
    pipeline based on a new peak picking method for EXploring Imaging
    Mass Spectrometry data." Bioinformatics 31.19 (2015): 3198-3206.
    https://academic.oup.com/bioinformatics/article/31/19/3198/212150
    """
    def __init__(self):
        super(EximsSelector, self).__init__()

    def _get_support_mask(self):
        """
        Get the boolean mask indicating which features are selected

        Returns
        -------
        support : boolean array of shape [# input features]
            An element is True iff its corresponding feature is selected for
            retention.
        """
        return self.selected_

    def fit(self, X, y=None, xy=None):
        """Learn data-driven feature thresholds from X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            Sample vectors from which to compute feature characteristic.

        y : any
            Ignored. This parameter exists only for compatibility with
            sklearn.pipeline.Pipeline.
        
        xy : array-like, shape (n_samples, 2)
            Spatial coordinates of the samples. Expects integers,
            indices over am image.

        Returns
        -------
        self
        """
        if xy is None:
            raise ValueError('xy coordinates are required')
        self.structness_ = exims(X, *xy.T)
        features_selection = select_features(self.structness_)
        self.threshold_ = features_selection.threshold
        self.selected_ = features_selection.selection
        return self



import unittest

import numpy as np
from sklearn.datasets import make_blobs
from sklearn.neighbors import KNeighborsClassifier

from divik.sampler._stratified_sampler import StratifiedSampler


def data():
    return make_blobs(n_samples=1000, n_features=30, centers=5, random_state=1)


class StratifiedSamplerTest(unittest.TestCase):
    def test_keeps_shapes(self):
        X, y = data()
        sampler = StratifiedSampler(n_rows=10).fit(X, y)
        self.assertEqual(sampler.get_sample(0).shape, (10, X.shape[1]))

    def test_keeps_proportion(self):
        X, y = data()
        sampler = StratifiedSampler().fit(X, y)
        X_rand = sampler.get_sample(0)
        # Rows should exist in X, so 1 neighbor is just exact match.
        y_rand = KNeighborsClassifier(n_neighbors=1).fit(X, y).predict(X_rand)
        _, orig_count = np.unique(y, return_counts=True)
        _, rand_count = np.unique(y_rand, return_counts=True)
        orig_prop = orig_count / np.linalg.norm(orig_count)
        rand_prop = rand_count / np.linalg.norm(rand_count)
        # This is equivalent to up to 1% of samples moved between labels
        self.assertLess(np.abs(orig_prop - rand_prop).max(), 0.01)

    def test_samples_exact_rows(self):
        X, y = data()
        sampler = StratifiedSampler().fit(X, y)
        X_rand = sampler.get_sample(0)
        for row in X_rand:
            self.assertTrue((row[np.newaxis, :] == X).all(axis=1).any(axis=0))


if __name__ == '__main__':
    unittest.main()



import divik._inspect.callback.visualization
import divik._inspect.callback.exclusion
import divik._inspect.callback.persistence
import divik._inspect.callback.recolor
import divik._inspect.callback.tabbing



"""Unsupervised feature selection methods"""
from ._stat_selector_mixin import (
    SelectorMixin,
    StatSelectorMixin,
    NoSelector,
)
from ._exims import EximsSelector
from ._gmm_selector import GMMSelector
from ._outlier import (
    huberta_outliers,
    OutlierSelector,
)
from ._percentage_selector import PercentageSelector
from ._specialized import (
    HighAbundanceAndVarianceSelector,
    OutlierAbundanceAndVarianceSelector,
    make_specialized_selector,
)


__all__ = [
    'SelectorMixin',
    'StatSelectorMixin',
    'NoSelector',
    'EximsSelector',
    'GMMSelector',
    'huberta_outliers',
    'OutlierSelector',
    'PercentageSelector',
    'HighAbundanceAndVarianceSelector',
    'OutlierAbundanceAndVarianceSelector',
    'make_specialized_selector',
]



import json

import numpy as np

from divik._summary import merged_partition
from divik._inspect.app import divik_result


def initialize_storage(level):
    return json.dumps({
        'excluded': [],
        'level': level,
        'reloaded_at': 0,
    })


def vec_in(v1, v2):
    if np.size(v2) == 0:
        return np.zeros((np.size(v1),), dtype=bool)
    return np.max(np.reshape(v1, (-1, 1)) == np.reshape(v2, (1, -1)), axis=1)


def update_excluded(old_level, new_level, old_excluded):
    # if any subcluster is enabled, parent cluster will be enabled
    # if parent is disabled, children will be disabled
    old_partition = merged_partition(divik_result(), old_level)
    old_enabled_flag = vec_in(old_partition, old_excluded) == 0

    new_partition = merged_partition(divik_result(), new_level)
    new_enabled_ids = np.unique(new_partition[old_enabled_flag])
    new_disabled_ids = np.setdiff1d(new_partition, new_enabled_ids)

    return sorted(list(map(int, new_disabled_ids)))


def update_storage(level, disabled_clusters, stamp, old_state):
    state = json.loads(old_state)
    state['excluded'] = disabled_clusters
    state['level'] = level
    state['reloaded_at'] = stamp
    return json.dumps(state)


def is_reloaded(stamp, old_state):
    state = json.loads(old_state)
    reloaded_at = state.get('reloaded_at', 0) or 0
    return stamp is not None and stamp > reloaded_at


def get_options(level):
    partition = merged_partition(divik_result(), level)
    return [{'label': e, 'value': e} for e in np.unique(partition)]


def update_actual_clusters(level, storage):
    if not storage:
        return []
    storage = json.loads(storage)
    if level == storage['level']:
        return storage['excluded']
    return update_excluded(storage['level'], level, storage['excluded'])



import unittest

import numpy as np
import numpy.testing as npt
from sklearn.base import clone

from divik.sampler._core import BaseSampler, ParallelSampler


class DummySampler(BaseSampler):
    def __init__(self, whatever):
        self.whatever = whatever

    def get_sample(self, seed):
        return np.array([[seed]])


class BaseSamplerTest(unittest.TestCase):
    def test_iterates_through_samples(self):
        sampler = DummySampler(1)
        samples = [v for _, v in zip(range(10000), sampler)]
        self.assertEqual(10000, len(samples))
        npt.assert_array_equal(np.ravel(samples), np.ravel(range(10000)))

    def test_fit_returns_self(self):
        sampler = DummySampler(1)
        self.assertIs(sampler, sampler.fit(np.array([[1]])))

    def test_parallel_sampler_generates_same_values(self):
        sampler = DummySampler(1)
        expected = sampler.get_sample(34134123)
        with sampler.parallel() as sampler_:
            actual = sampler_.get_sample(34134123)
            npt.assert_array_equal(expected, actual)

    def test_context_creates_parallel_sampler(self):
        sampler = DummySampler(1)
        with sampler.parallel() as sampler_:
            self.assertIsInstance(sampler_, ParallelSampler)

    def test_parallel_sampler_is_cloneable(self):
        sampler = DummySampler(1)
        with sampler.parallel() as sampler_:
            cloned = sampler_.clone()
            self.assertIsInstance(cloned, DummySampler)
            self.assertEqual(sampler.whatever, cloned.whatever)
    
    def test_parallel_sampler_is_not_iterable(self):
        sampler = DummySampler(1)
        with sampler.parallel() as sampler_, self.assertRaises(TypeError):
            iter(sampler_)
    
    def test_parallel_sampler_is_not_fittable(self):
        sampler = DummySampler(1)
        with sampler.parallel() as sampler_, self.assertRaises(AttributeError):
            sampler_.fit()

    def test_is_cloneable(self):
        original = DummySampler(1)
        cloned = clone(original)
        self.assertEqual(cloned.whatever, 1)



import unittest
from unittest.mock import create_autospec, patch

import numpy as np

from divik.cluster._kmeans import _initialization as km
from test.cluster.kmeans import data


def measure(func):
    return create_autospec(func, side_effect=func)


class ExtremeInitializationTest(unittest.TestCase):
    def setUp(self):
        self.number_of_clusters = 2
        self.initialize = km.ExtremeInitialization('euclidean')

    def test_centroids_have_the_same_number_of_features_as_data(self):
        centroids = self.initialize(data, self.number_of_clusters)
        self.assertEqual(centroids.shape[1], data.shape[1])

    def test_number_of_centroids_is_preserved(self):
        centroids = self.initialize(data, self.number_of_clusters)
        self.assertEqual(centroids.shape[0], self.number_of_clusters)

    def test_works_without_error_for_ill_conditioned_problems(self):
        self.initialize(data[0:3], self.number_of_clusters)

    def test_throws_when_required_more_centroids_than_data(self):
        with self.assertRaises(ValueError):
            self.initialize(data[0:self.number_of_clusters-1],
                            self.number_of_clusters)

    def test_first_centroid_is_furthest_from_all(self):
        centroid = self.initialize(data, 1)
        np.testing.assert_equal(centroid, data[-1].reshape(1, -1))

    def test_next_centroid_is_furthest_from_already_found(self):
        centroids = self.initialize(data, 3)
        np.testing.assert_equal(centroids[1], data[4])
        np.testing.assert_equal(centroids[2], data[2])


# 99-th percentile should be approximated to max
class PercentileInitializationTest(ExtremeInitializationTest):
    def setUp(self):
        self.number_of_clusters = 2
        self.initialize = km.PercentileInitialization('euclidean')


class KDTreePercentileInitializationTest(unittest.TestCase):
    def setUp(self):
        self.number_of_clusters = 2
        self.initialize = km.KDTreePercentileInitialization(
            'euclidean', 0.1, 99.)

    def test_works_for_sample_data(self):
        self.initialize(data, self.number_of_clusters)



from dash.dependencies import Output, Input, State

from divik._inspect import exclusion as ex, persistence as per
from divik._inspect.app import app
from divik._inspect.layout import Fields


@app.callback(
    Output(Fields.DISABLED_CLUSTERS_STORAGE, 'children'),
    [
        Input(Fields.DISABLED_CLUSTERS_PICKER, 'value'),
    ],
    [
        State(Fields.LOAD_PROFILE, 'n_clicks_timestamp'),
        State(Fields.LEVEL, 'value'),
        State(Fields.DISABLED_CLUSTERS_STORAGE, 'children'),
    ]
)
def update_disabled_clusters_to_new_level(disabled_clusters, stamp, level,
                                          old_state):
    if not old_state:
        return ex.initialize_storage(level)
    return ex.update_storage(level, disabled_clusters, stamp, old_state)


@app.callback(
    Output(Fields.DISABLED_CLUSTERS_PICKER, 'options'),
    [Input(Fields.LEVEL, 'value')]
)
def update_possible_enabled_clusters(level):
    return ex.get_options(level)


@app.callback(
    Output(Fields.DISABLED_CLUSTERS_PICKER, 'value'),
    [Input(Fields.LEVEL, 'value')],
    [
        State(Fields.DISABLED_CLUSTERS_STORAGE, 'children'),
        State(Fields.LOAD_PROFILE, 'n_clicks_timestamp'),
        State(Fields.SAVED_PROFILES, 'value'),
    ]
)
def update_actually_disabled_clusters(level, storage, stamp, name):
    if not storage:
        return []
    if ex.is_reloaded(stamp, storage):
        return per.restore_disabled_clusters(name)
    return ex.update_actual_clusters(level, storage)



import unittest

import numpy as np
from scipy.spatial.distance import cdist
from sklearn.datasets import make_blobs

from divik.feature_extraction import KneePCA


def data():
    return make_blobs(n_samples=100, n_features=30, centers=5, random_state=1)


class KneePCATest(unittest.TestCase):
    def test_fits_with_no_error(self):
        X, _ = data()
        KneePCA().fit(X)

    def test_fit_transform_limits_dims(self):
        X, _ = data()
        X_trans = KneePCA().fit_transform(X)
        self.assertLess(X_trans.shape[1], X.shape[1])

    def test_fit_transform_preserves_all_observations(self):
        X, _ = data()
        X_trans = KneePCA().fit_transform(X)
        self.assertEqual(X.shape[0], X_trans.shape[0])

    def test_inverse_transform_preserves_shape(self):
        X, _ = data()
        pca = KneePCA()
        X_trans = pca.fit_transform(X)
        X_inv = pca.inverse_transform(X_trans)
        self.assertEqual(X_inv.shape, X.shape)

    def test_inverse_transform_preserves_most_values(self):
        X, _ = data()
        pca = KneePCA()
        X_trans = pca.fit_transform(X)
        X_inv = pca.inverse_transform(X_trans)
        # rows should be at least correlated a bit
        d = cdist(X_inv, X, metric='correlation')
        d = d[np.eye(d.shape[0], dtype=bool)]
        self.assertLess(d.max(), 0.2)
    
    def test_there_is_not_much_data_loss(self):
        X, _ = make_blobs(n_samples=100, n_features=3, centers=5,
                          random_state=0)
        pca = KneePCA()
        X_trans = pca.fit_transform(X)
        X_inv = pca.inverse_transform(X_trans)
        np.testing.assert_array_almost_equal(X, X_inv, decimal=3)

    def test_refit_inverse_transform_preserves_shape(self):
        X, _ = data()
        pca = KneePCA(refit=True)
        X_trans = pca.fit_transform(X)
        X_inv = pca.inverse_transform(X_trans)
        self.assertEqual(X_inv.shape, X.shape)

    def test_refit_does_not_introduce_data_loss(self):
        X, _ = make_blobs(n_samples=100, n_features=3, centers=5,
                          random_state=0)
        pca = KneePCA(refit=True)
        X_trans = pca.fit_transform(X)
        X_inv = pca.inverse_transform(X_trans)
        np.testing.assert_array_almost_equal(X, X_inv, decimal=3)


if __name__ == '__main__':
    unittest.main()
